xSized bit strings, an example of an indexed type, and some further
associated types.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Bits where

open import VFP.Natural.Non-Strict-Order

private variable
  i j k m n : ℕ
\end{code}

-------------------------------------------------------------------------------

Sized bit strings.

<<Bits>>
\begin{code}
data Bits : ℕ -> Set where
  []   :            Bits (zero)
  𝟢 𝟣  : Bits n ->  Bits (succ n)
\end{code}

\begin{code}
Eau-de-Cologne = 𝟣 (𝟣 (𝟣 (𝟢 (𝟢 (𝟣 (𝟣 (𝟢 (𝟢 (𝟣 (𝟢 (𝟢 (𝟣 []))))))))))))
\end{code}

Bits n is downwards closed i.e. it is an object of the functor
category ℕ → Set.

<<take>>
\begin{code}
take : (n ≽ m) -> (Bits n -> Bits m)
take (zero)      bs      = []
take (succ n≽m)  (𝟢 bs)  = 𝟢 (take n≽m bs)
take (succ n≽m)  (𝟣 bs)  = 𝟣 (take n≽m bs)
\end{code}

The function take is the functorial action of Bits on arrows. Proof of
the functor laws.

\begin{code}
take-preserves-identity : ∀ (bs : Bits n) ->
    take ≽-reflexive bs ≡ bs
take-preserves-identity {zero}   []     = ≡-reflexive
take-preserves-identity {succ n} (𝟢 bs) = congruent 𝟢 (take-preserves-identity bs)
take-preserves-identity {succ n} (𝟣 bs) = congruent 𝟣 (take-preserves-identity bs)

take-preserves-composition : ∀ (n≽m : n ≽ m) (m≽k : m ≽ k) (bs : Bits n) ->
    take (≽-transitive n≽m m≽k) bs ≡ take m≽k (take n≽m bs)
take-preserves-composition zero zero bs = ≡-reflexive
take-preserves-composition (succ n≽m) (zero)     bs     = ≡-reflexive
take-preserves-composition (succ n≽m) (succ m≽k) (𝟢 bs) = congruent 𝟢 (take-preserves-composition n≽m m≽k bs)
take-preserves-composition (succ n≽m) (succ m≽k) (𝟣 bs) = congruent 𝟣 (take-preserves-composition n≽m m≽k bs)
\end{code}

NB. take is a covariant functor (indeed, the notation has been
carefully chosen so that this is the case); that the order of
composition is reversed in take-preserves-composition is due to the
fact that ≽-transitive implements forward composition whereas ∘
(implictly used on the RHS) is backward composition of functions.

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Bit lists of size *below* an upper bound.

\begin{code}
infixr 6 _such-that_
\end{code}

<<Bits-below>>
\begin{code}
record Bits≼ (n : ℕ) : Set where
  constructor _such-that_
  field  {size} : ℕ
         bits   : Bits size
         below  : n ≽ size
\end{code}

\begin{code}
open Bits≼ public
\end{code}

Bits≼ n is *not* downwards closed, it is upwards closed (duality at
work) e.g. we can raise the upper bound.

<<raise>>
\begin{code}
raise : (i ≽ j) -> (Bits≼ j -> Bits≼ i)
raise i≽j (bs such-that j≽k) = bs such-that ≽-transitive i≽j j≽k
\end{code}

NB. raise only works on the proof, the actual data, the bit string, is
unchanged.

\begin{code}
raise-preserves-identity : ∀ (bs : Bits≼ n) ->
    raise ≽-reflexive bs ≡ bs
raise-preserves-identity (bs such-that n≽m) = congruent (bs such-that_) (≽-left-identity n≽m)    

raise-preserves-composition : ∀ (n≽m : n ≽ m) (m≽k : m ≽ k) (bs : Bits≼ k) ->
    raise (≽-transitive n≽m m≽k) bs ≡ raise n≽m (raise m≽k bs)
raise-preserves-composition n≽m m≽k (bs such-that k≽j) = congruent (bs such-that_) (≽-associative n≽m m≽k k≽j)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Bit lists of size *above* a lower bound.

<<Bits-above>>
\begin{code}
record Bits≽ (n : ℕ) : Set where
  constructor _such-that_
  field  {size}  : ℕ
         bits    : Bits size
         above   : n ≼ size
\end{code}

\begin{code}
open Bits≽ public
\end{code}

-------------------------------------------------------------------------------
