Boolean values: false and true.

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Boolean where

open import VFP.Logic
open import VFP.Equality
open import VFP.Bijection
\end{code}

-------------------------------------------------------------------------------

<<Boolean>>
\begin{code}
data Boolean : Set where
  false  : Boolean
  true   : Boolean
\end{code}

\begin{code}
{-# BUILTIN BOOL  Boolean  #-}
{-# BUILTIN FALSE false #-}
{-# BUILTIN TRUE  true  #-}
\end{code}

<<Boolean-mathbb>>
\begin{code}
𝔹 : Set
𝔹 = Boolean
\end{code}

<<conditional>>
\begin{code}
infix 0 if_then_else_
if_then_else_ : {A : Set} -> Boolean → A → A → A
if false  then t else e = e
if true   then t else e = t
\end{code}

See also

infix 0 if_then_else_
if_then_else_ : {A : Set} -> (b : Boolean) → ({So b} -> A) → ({So (not b)}-> A) → A
if false  then t else e = e {yes}
if true   then t else e = t {yes}

in VFP.Searching.Binary-Search.Haskell.

The free theorem of „if_then_else_“ — the relation R is an explicit
argument to avoid unsolved meta-variables.

<<at-if>>
\begin{code}
at_if_then_else_ : ∀ {A A′ : Set} (R : A -> A′ -> Set)
                 → ∀ {b b′} -> b ≡ b′
                 → ∀ {a a′} -> R a a′
                 → ∀ {c c′} -> R c c′
                 → R (if b then a else c) (if b′ then a′ else c′)
at_if_then_else_ R {b = true } reflexive q r = q
at_if_then_else_ R {b = false} reflexive q r = r
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Logical connectives.

<<not>>
\begin{code}
not : 𝔹 -> 𝔹

not false = true
not true  = false
\end{code}

\begin{code}
infixr  9 _&&_
infixr  8 _||_
infixr  7 _=>_ _<=>_

_&&_ _||_ _=>_ _<=>_ : 𝔹 -> 𝔹 -> 𝔹

false && b = false
true  && b = b

false || b = b
true  || b = true

false => b = true
true  => b = b

false <=> false = true
false <=> true  = false
true  <=> false = false
true  <=> true  = true
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Interdefinability of logical connectives (using false and =>).

\begin{code}
def-not : ∀ b -> b => false ≡ not b
def-not false = reflexive
def-not true  = reflexive

def-|| : ∀ b₁ b₂ -> (b₂ => false) => b₁ ≡ b₁ || b₂
def-|| false false = reflexive
def-|| true  false = reflexive
def-|| false true  = reflexive
def-|| true  true  = reflexive

def-&& : ∀ b₁ b₂ -> (b₁ => b₂ => false) => false ≡ b₁ && b₂
def-&& false b₂   = reflexive
def-&& true false = reflexive
def-&& true true  = reflexive
\end{code}

-------------------------------------------------------------------------------

When is a Boolean expression true? This is „so true“.

<<So>>
\begin{code}
data So : Boolean -> Set where
  yes : So true
\end{code}

|So b| is equivalent to |b ≡ true|.

true:

\begin{code}
true-intro : So true
true-intro = yes
\end{code}

\begin{code}
true-to : ⊤ -> So true
true-to tt = yes

true-fro : So true -> ⊤
true-fro yes = trivial
\end{code}

false:

\begin{code}
false-elim : ∀ {b : Boolean} -> So false -> So b
false-elim ()
\end{code}

\begin{code}
false-to : ⊥ -> So false
false-to ()

false-fro : So false -> ⊥
false-fro ()
\end{code}

Negation:

\begin{code}
not-intro : ∀ {b : Boolean} -> So (b => false) -> So (not b)
not-intro {false} yes = yes

not-elim : ∀ {b : Boolean} -> So b -> So (not b) -> So false
not-elim yes ()
\end{code}

\begin{code}
not-to : ∀ {b : Boolean} -> ¬ So b -> So (not b)
not-to {false} f = yes
not-to {true}  f = ex-falso-quodlibet (f yes)

not-fro : ∀ {b : Boolean} -> So (not b) -> ¬ So b
not-fro {false} yes = λ ()
\end{code}

Implication:

<<intro>>
\begin{code}
=>-intro : ∀ {b₁ b₂ : Boolean} -> (So b₁ -> So b₂) -> So (b₁ => b₂)
=>-intro {false} f = yes
=>-intro {true}  f = f yes
\end{code}

<<elim>>
\begin{code}
=>-elim : ∀ {b₁ b₂ : Boolean} -> So (b₁ => b₂) -> (So b₁ -> So b₂)
=>-elim yes yes = yes
\end{code}

\begin{code}
xx : ∀ {b₁ b₂ : Boolean} -> ∀ (so : So (b₁ => b₂)) -> =>-intro {b₁ = b₁} (=>-elim so) ≡ so
xx {false} yes = reflexive
xx {true}  yes = reflexive

yy : ∀ {b₁ b₂ : Boolean} -> ∀ (f : So b₁ -> So b₂) -> ∀ so → =>-elim (=>-intro f) so ≡ f so
yy f yes with f yes
... | yes = reflexive
\end{code}

Conjunction:

\begin{code}
&&-intro : ∀ {b₁ b₂ : Boolean} -> So b₁ -> So b₂ -> So (b₁ && b₂)
&&-intro yes yes = yes

&&-elim₁ : ∀ {b₁ b₂ : Boolean} -> So (b₁ && b₂) -> So b₁
&&-elim₁ {true} yes = yes

&&-elim₂ : ∀ {b₁ b₂ : Boolean} -> So (b₁ && b₂) -> So b₂
&&-elim₂ {true} yes = yes
\end{code}

\begin{code}
&&-to : ∀ {b₁ b₂ : Boolean} -> So b₁ × So b₂ -> So (b₁ && b₂)
&&-to (so-b₁ , so-b₂) = &&-intro so-b₁ so-b₂

&&-fro : ∀ {b₁ b₂ : Boolean} -> So (b₁ && b₂) -> So b₁ × So b₂
&&-fro so = &&-elim₁ so , &&-elim₂ so

&&-to-fro : ∀ {b₁ b₂ : Boolean} -> ∀ (x : So (b₁ && b₂)) -> &&-to {b₁ = b₁} (&&-fro x) ≡ x
&&-to-fro {true} {true} yes = reflexive

&&-fro-to : ∀ {b₁ b₂ : Boolean} -> ∀ (x : So b₁ × So b₂) -> &&-fro {b₁ = b₁} (&&-to x) ≡ x
&&-fro-to (yes , yes) = reflexive

&&-iso : ∀ {b₁ b₂ : Boolean} -> So (b₁ && b₂) ≅ So b₁ × So b₂
&&-iso = record
  { to  = &&-fro
  ; fro = &&-to
  ; fro∘to≡id = &&-to-fro
  ; to∘fro≡id = &&-fro-to }
\end{code}

Disjunction:

\begin{code}
||-intro₁ : ∀ {b₁ b₂ : Boolean} -> So b₁ -> So (b₁ || b₂)
||-intro₁ yes = yes

||-intro₂ : ∀ {b₁ b₂ : Boolean} -> So b₂ -> So (b₁ || b₂)
||-intro₂ {false} yes = yes
||-intro₂ {true}  yes = yes

||-elim : ∀ {b₁ b₂ b₃ : Boolean} -> So (b₁ || b₂) -> So (b₁ => b₃) -> So (b₂ => b₃) -> So b₃
||-elim {false} {.true} yes B C = C
||-elim {true} {b₂}     yes B C = B
\end{code}

\begin{code}
||-to : ∀ {b₁ b₂ : Boolean} -> So b₁ ⊎ So b₂ -> So (b₁ || b₂)
||-to {false} {true}  (inr yes) = yes
||-to {true}  {false} (inl yes) = yes
||-to {true}  {true}  (inl yes) = yes
||-to {true}  {true}  (inr yes) = yes

||-fro : ∀ {b₁ b₂ : Boolean} -> So (b₁ || b₂) -> So b₁ ⊎ So b₂
||-fro {false} {true}  yes = inr yes
||-fro {true}  {false} yes = inl yes
||-fro {true}  {true}  yes = inl yes -- OR: inr yes

||-to-fro : ∀ {b₁ b₂ : Boolean} -> ∀ (x : So (b₁ || b₂)) -> ||-to {b₁ = b₁} (||-fro {b₁ = b₁} x) ≡ x
||-to-fro {false} {true}  yes = reflexive
||-to-fro {true}  {false} yes = reflexive
||-to-fro {true}  {true}  yes = reflexive
\end{code}

The converse is not true. ||-to is not injective: it maps inl yes and
inr yes to yes (b₁ = true and b₁ = true).

||-fro-to : ∀ {b₁ b₂ : Boolean} -> ∀ (x : So b₁ ⊎ So b₂) -> ||-fro {b₁ = b₁} (||-to {b₁ = b₁} x) ≡ x
||-fro-to {false} {true}  (inr yes) = reflexive
||-fro-to {true}  {false} (inl yes) = reflexive
||-fro-to {true}  {true}  (inl yes) = reflexive
||-fro-to {true}  {true}  (inr yes) = {!!}

Excluded middle:

\begin{code}
Boolean-excluded-middle : ∀ (b : Boolean) -> So (b || not b)
Boolean-excluded-middle false = yes
Boolean-excluded-middle true  = yes
\end{code}

<<double-negation>>
\begin{code}
Boolean-double-negation : ∀ (b : Boolean) -> So (((b => false) => false) => b)
Boolean-double-negation false = yes
Boolean-double-negation true  = yes
\end{code}

-------------------------------------------------------------------------------
