Standard cons lists.

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.List where

open import VFP.Boolean
open import VFP.Maybe hiding (map ; return ; _>>=_)
open import VFP.Natural
open import VFP.Decidable

private variable
  A Elem : Set
\end{code}

-------------------------------------------------------------------------------

NB. If the flag --without-K is enabled then

data List : Set -> Set where
  []  : List Elem
  _∷_ : Elem -> List Elem -> List Elem

gives the error message: Set₁ is not less or equal than Set.
The variant below, however, works.

<<cons-fixity>>
\begin{code}
infixr 7 _∷_
\end{code}

<<List>>
\begin{code}
data List (Elem : Set) : Set where
  []   : List Elem
  _∷_  : Elem -> List Elem -> List Elem
\end{code}

\begin{code}
𝕃 = List
\end{code}

\begin{code}
{-# BUILTIN LIST List #-}
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
null : List A -> Boolean
null []       = true
null (a ∷ as) = false
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
map : ∀ {A B : Set} -> (A -> B) -> (List A -> List B)
map f []        = []
map f (a ∷ as)  = f a ∷ map f as
\end{code}

\begin{code}
infix 8 [_]

[_] : ∀ {Elem : Set} -> Elem → List Elem
[ x ] = x ∷ []

return : ∀ {A : Set} -> A -> List A
return a = a ∷ []
\end{code}

<<append>>
\begin{code}
infixr 7 _⧺_
_⧺_ : ∀ {Elem : Set} -> List Elem → List Elem → List Elem
[]       ⧺ ys = ys
(x ∷ xs) ⧺ ys = x ∷ (xs ⧺ ys)
\end{code}

\begin{code}
concat : ∀ {A : Set} -> List (List A) -> List A
concat []         = []
concat (as ∷ aas) = as ⧺ concat aas
\end{code}

<<length>>
\begin{code}
length : ∀ {Elem : Set} → List Elem → ℕ
length []       = zero
length (x ∷ xs) = succ (length xs)
\end{code}
length []       = 0
length (x ∷ xs) = 1 + length xs

<<foldr>>
\begin{code}
foldr : {P S : Set} -> (P -> S -> S) -> S -> List P -> S
foldr _◁_ s []       = s
foldr _◁_ s (x ∷ xs) = x ◁ (foldr _◁_ s xs)
\end{code}

\begin{code}
foldl : {P S : Set} -> (S -> P -> S) -> S -> List P -> S
foldl _▷_ s []       = s
foldl _▷_ s (x ∷ xs) = foldl _▷_ (s ▷ x) xs
\end{code}

\begin{code}
sum : List ℕ -> ℕ
sum = foldr _+_ 0
\end{code}

\begin{code}
and or : List 𝔹 -> 𝔹
and = foldr _&&_ true
or  = foldr _||_ false
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
mapMaybe : ∀ {A B : Set} -> (A -> Maybe B) -> (List A -> List B)
mapMaybe f [] = []
mapMaybe f (a ∷ as) with f a
... | Nothing =     mapMaybe f as
... | Just b  = b ∷ mapMaybe f as
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
filter : ∀ {P : A -> Set} → Decidable P → List A → List A
filter P? [] = []
filter P? (x ∷ xs) with P? x
... | yes p = x ∷ filter P? xs
... | no ¬p =     filter P? xs
\end{code}

-------------------------------------------------------------------------------

\begin{code}
slow-reverse : ∀ {Elem : Set} → List Elem → List Elem
slow-reverse []       = []
slow-reverse (x ∷ xs) = slow-reverse xs ⧺ (x ∷ [])

reverse = slow-reverse

reverse-cat : ∀ {Elem : Set} → List Elem → List Elem → List Elem
reverse-cat []       ys = ys
reverse-cat (x ∷ xs) ys = reverse-cat xs (x ∷ ys)

fast-reverse : ∀ {Elem : Set} → List Elem → List Elem
fast-reverse xs = reverse-cat xs []
\end{code}

\begin{code}
open import VFP.Natural.Order

lookup : (xs : List Elem) -> {n : ℕ} -> n ≺ length xs -> Elem
lookup (x ∷ xs) (zero≺succ)   = x
lookup (x ∷ xs) (succ≺succ p) = lookup xs p
\end{code}

-------------------------------------------------------------------------------

Predicates i.e. Set-valued functions. (forall is a keyword, which is
why the universal predicate is called all not forall. Haskell defines
all , any :: (a -> Bool) -> ([a] -> Bool).)

<<all>>
\begin{code}
data all (P : Elem -> Set) : (List Elem -> Set) where
  []   :                                 all P []
  _∷_  : ∀ {a as} -> P a -> all P as ->  all P (a ∷ as)

zip : {P : Elem -> Set} -> (xs : List Elem) -> all P xs -> List (Σ Elem P)
zip []       []       = []
zip (x ∷ xs) (p ∷ ps) = (x , p) ∷ zip xs ps
\end{code}

NB. The constructors [] and _∷_ are overloaded.

The empty list trivially satisfies the property |all P|; a non-empty
list satisfies the property |all P| if the head and each element of
the tail satisfies the property |P|.

<<exists>>
\begin{code}
data exists (P : Elem -> Set) : (List Elem -> Set) where
  head  : ∀ {a as} -> P a          -> exists P (a ∷ as)
  tail  : ∀ {a as} -> exists P as  -> exists P (a ∷ as)
\end{code}

A non-empty list satisfies the property |exists P| if either the head
satisfies |P| or any element in the tail (the constructors head and
tail can be seen as paths).

Membership.

\begin{code}
open import VFP.Equality
\end{code}

<<membership>>
\begin{code}
infix 6 _∈_
_∈_ : A -> List A -> Set
a ∈ xs = exists (λ x -> a ≡ x) xs
\end{code}

-------------------------------------------------------------------------------

Views (see Idris §10).

\begin{code}
data Rear {Elem : Set} : List Elem -> Set where
  empty     : Rear []
  non-empty : ∀ xs x -> Rear (xs ⧺ [ x ])

rear : (xs : List Elem) -> Rear xs
rear [] = empty
rear (x ∷ xs) with rear xs
... | empty            = non-empty [] x
... | non-empty xs′ x′ = non-empty (x ∷ xs′) x′

reverse′ : List Elem → List Elem
reverse′ xs with rear xs
reverse′ .[]           | empty          = []
reverse′ .(xs ⧺ [ x ]) | non-empty xs x = x ∷ reverse xs
\end{code}

The dots cannot be removed as xs ⧺ [ x ] is not a legal pattern.

-------------------------------------------------------------------------------
