Natural numbers: unary representation aka Peano numerals.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural where

open import VFP.Boolean
\end{code}

-------------------------------------------------------------------------------

<<Natural>>
\begin{code}
data Natural : Set where
  zero  :  Natural
  succ  :  Natural → Natural
\end{code}

<<blackboard>>
\begin{code}
ℕ : Set
ℕ = Natural
\end{code}

\begin{code}
{-# BUILTIN NATURAL Natural #-}
\end{code}

The |BUILTIN| pragma enables us to use |0|, |1| etc as naturals.

Folding natural numbers.

<<recurse>>
\begin{code}
recurse : ∀ {A : Set} -> (A -> A) -> A -> ℕ -> A
recurse s z (zero)    = z
recurse s z (succ n)  = s (recurse s z n)
\end{code}

<<iterate>>
\begin{code}
iterate : ∀ {A : Set} -> (A -> A) -> A -> (ℕ -> A)
iterate s z (zero)    = z
iterate s z (succ n)  = iterate s (s z) n
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Basic arithmetic.

\begin{code}
pred : ℕ -> ℕ
pred (zero)   = zero
pred (succ n) = n
\end{code}

<<plus>>
\begin{code}
infixl 8 _+_
_+_ : Natural -> Natural -> Natural
zero    + n = n
succ m  + n = succ (m + n)
\end{code}

<<plus-reverse>>
\begin{code}
infixl 8 _+ᴿ_
_+ᴿ_ : Natural -> Natural -> Natural
m +ᴿ zero   = m
m +ᴿ succ n = succ (m +ᴿ n)
\end{code}

\begin{code}
infixl 8 _∸_
_∸_ : Natural -> Natural -> Natural
zero   ∸ n      = zero
succ m ∸ zero   = succ m
succ m ∸ succ n = m ∸ n

infixl 9 _*_
_*_ : Natural -> Natural -> Natural
zero   * n = zero
succ m * n = n + m * n

--_·2 : Peano -> Peano
--zero    ·2 = zero
--succ n  ·2 = succ (succ (n ·2))

infixl 9 2*_
2*_ : Natural -> Natural
2* zero   = zero
2* succ n = succ (succ (2* n))

div2 : ℕ -> ℕ
div2 zero = zero
div2 (succ zero) = zero
div2 (succ (succ n)) = succ (div2 n)
\end{code}

Naïve exponentiation.

\begin{code}
_^_ : ℕ → ℕ → ℕ
x ^ zero   = 1
x ^ succ n = x * x ^ n

2^_ : ℕ → ℕ
2^ zero   = 1
2^ succ n = 2* 2^ n
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Division by 2, rounded downwards and upwards.

\begin{code}
⌊_/2⌋ : ℕ → ℕ
⌊ 0 /2⌋             = 0
⌊ 1 /2⌋             = 0
⌊ succ (succ n) /2⌋ = succ ⌊ n /2⌋

⌈_/2⌉ : ℕ → ℕ
⌈ n /2⌉ = ⌊ succ n /2⌋
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Comparison.

<<ordering>>
\begin{code}
infix 6 _≤_
_≤_ : Natural -> Natural -> Boolean

zero    ≤ n       =  true
succ m  ≤ zero    =  false
succ m  ≤ succ n  =  m ≤ n
\end{code}

\begin{code}
infix 6 _≥_
_≥_ : Natural -> Natural -> Boolean
m ≥ n = n ≤ m
\end{code}

m < n = not (m ≥ n)

<<strict-ordering>>
\begin{code}
infix 6 _<_
_<_ : Natural -> Natural -> Boolean
m       < zero    = false
zero    < succ n  = true
succ m  < succ n  = m < n
\end{code}

\begin{code}
infix 6 _>_
_>_ : Natural -> Natural -> Boolean
m > n = not (m ≤ n)
\end{code}

Minimum and maximum (Order defines ↑).

\begin{code}
_⊔_ : ℕ → ℕ → ℕ
zero  ⊔ n       = n
succ m ⊔ zero   = succ m
succ m ⊔ succ n = succ (m ⊔ n)

_⊓_ : ℕ → ℕ → ℕ
zero  ⊓ n       = zero
succ m ⊓ zero   = zero
succ m ⊓ succ n = succ (m ⊓ n)
\end{code}

-------------------------------------------------------------------------------

Unicode: ∸ \.-

-------------------------------------------------------------------------------
