Non-strict ordering on the naturals (defines ≽).

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural.Non-Strict-Order where

open import VFP.Natural.Properties public

private variable
  j k m n : ℕ
\end{code}

-------------------------------------------------------------------------------

Non-strict ordering on ℕ, n ≽ m (pronounce: n succeeds or equals m)
and derived strict ordering, n ≻ m (pronounce: n succeeds m).

\begin{code}
infix 6 _≽_ _≻_
\end{code}

<<succeeds>>
\begin{code}
data _≽_ : ℕ → ℕ → Set where
    zero  : n ≽ zero
    succ  : n ≽ m → succ n ≽ succ m
\end{code}

\begin{code}
_≻_ : ℕ -> ℕ -> Set
n ≻ m = n ≽ succ m

infix 6 _≼_ _≺_
_≼_ _≺_ : ℕ -> ℕ -> Set
m ≼ n = n ≽ m
m ≺ n = n ≻ m
\end{code}

NB. The proof zero captures that zero (the first natural) is the least
element; the proof succ captures that succ (the successor function) is
monotone. Also observe that the proof of n ≽ m has the same shape as
the natural m itself.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Properties of the non-strict order |≽|.

<<signatures>>
\begin{code}
≽-reflexive   : n ≽ n
≽-transitive  : n ≽ m → m ≽ k → n ≽ k
succ-n≽n      : succ n ≽ n
\end{code}

\begin{code}
≽-reflexive {n = zero  } = zero
≽-reflexive {n = succ n} = succ ≽-reflexive

≽-antisymmetric : n ≽ m → m ≽ n → m ≡ n
≽-antisymmetric (zero)     (zero)      = ≡-reflexive
≽-antisymmetric (succ n≽m) (succ m≽n) = ≡-congruent succ (≽-antisymmetric n≽m m≽n)

≽-transitive n≽m        (zero)     = zero
≽-transitive (succ n≽m) (succ m≽k) = succ (≽-transitive n≽m m≽k)
\end{code}

≽-transitive (zero)     (zero)     = zero
≽-transitive (succ n≽m) (zero)     = zero
≽-transitive (succ n≽m) (succ m≽k) = succ (≽-transitive n≽m m≽k)

Relating ordering and arithmetic.

\begin{code}
succ-n≽n {n = zero  } = zero
succ-n≽n {n = succ n} = succ succ-n≽n
\end{code}

NB. The proofs of ≽-reflexive and succ-n≽n are identical; they really
show that n + k ≽ n.

\begin{code}
n+k≽n : n + k ≽ n
n+k≽n {n = zero  } = zero
n+k≽n {n = succ n} = succ n+k≽n

≽-reflexive′ : n ≽ n
≽-reflexive′ {n} = substitutive (λ hole -> hole ≽ n) +-right-identity n+k≽n

succ-n≽n′ : succ n ≽ n
succ-n≽n′ {n} = substitutive (λ hole -> hole ≽ n) (transitive (shift n 0) +-right-identity) n+k≽n
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

A partial order such as (ℕ, ≽) forms a category: ≽-reflexive is the
identity and ≽-transitive is composition. Composition is associative
with identity as its neutral element (properties of properties).

\begin{code}
≽-left-identity : ∀ (n≽m : n ≽ m) -> ≽-transitive ≽-reflexive n≽m ≡ n≽m
≽-left-identity (zero)     = reflexive
≽-left-identity (succ n≽m) = congruent succ (≽-left-identity n≽m)

≽-right-identity : ∀ (n≽m : n ≽ m) -> ≽-transitive n≽m ≽-reflexive ≡ n≽m
≽-right-identity (zero)     = reflexive
≽-right-identity (succ n≽m) = congruent succ (≽-right-identity n≽m)

≽-associative : ∀ (n≽m : n ≽ m) (m≽k : m ≽ k) (k≽j : k ≽ j) ->
    ≽-transitive (≽-transitive n≽m m≽k) k≽j ≡ ≽-transitive n≽m (≽-transitive m≽k k≽j)
≽-associative (zero)     (zero)     (zero)     = reflexive
≽-associative (succ n≽m) (zero)     (zero)     = reflexive
≽-associative (succ n≽m) (succ m≽k) (zero)     = reflexive
≽-associative (succ n≽m) (succ m≽k) (succ k≽j) = congruent succ (≽-associative n≽m m≽k k≽j)
\end{code}

All of these properties follow also from the fact that there is at
most one arrow between any two objects: ≽-proofs are unique.

\begin{code}
≽-unique : (p : n ≽ m) -> (q : n ≽ m) -> p ≡ q
≽-unique (zero)   (zero)   = reflexive
≽-unique (succ p) (succ q) = congruent succ (≽-unique p q)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Renamings:

\begin{code}
≼-reflexive : n ≼ n
≼-reflexive = ≽-reflexive

≼-transitive : k ≼ m -> m ≼ n -> k ≼ n
≼-transitive q p = ≽-transitive p q

n≼succ-n = succ-n≽n
\end{code}

-------------------------------------------------------------------------------

