% Unicode: ≺ \prec ; ≻ \succ ; ≼ \preceq ; ≽ \succeq

\begin{code}
{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --safe #-}

module VFP.Natural.Order.Properties where

open import VFP.Natural.Order public
open import VFP.Equality.Properties public
open import VFP.Logic.Properties
open import VFP.Decidable

private variable
  k m n : ℕ
\end{code}

-------------------------------------------------------------------------------

Properties of the strict order |≺|.

<<irreflexive>>
\begin{code}
≺-irreflexive : ¬ n ≺ n
≺-irreflexive (succ≺succ n≺n) = ≺-irreflexive n≺n
\end{code}

<<transitive>>
\begin{code}
≺-transitive : k ≺ m → m ≺ n → k ≺ n
≺-transitive (zero≺succ)     (succ≺succ _  ) = zero≺succ
≺-transitive (succ≺succ k≺m) (succ≺succ m≺n) = succ≺succ (≺-transitive k≺m m≺n)
\end{code}

Uniqueness of proofs.

\begin{code}
≺-unique : ∀ {m n} -> (p : m ≺ n) -> (q : m ≺ n) -> p ≡ q
≺-unique (zero≺succ)   (zero≺succ)                        = reflexive
≺-unique (succ≺succ p) (succ≺succ q) rewrite ≺-unique p q = reflexive
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Private stuff for lecture slides and sessions.

\begin{code}
private
\end{code}

<<3-precedes-5>>
\begin{code}
  three-precedes-five : 3 ≺ 5
  three-precedes-five = succ≺succ (succ≺succ (succ≺succ zero≺succ))
\end{code}

<<transitive-private>>
\begin{code}
  trans : k ≺ m → m ≺ n → k ≺ n
  trans (zero≺succ)      (succ≺succ _  )  = zero≺succ
  trans (succ≺succ k≺m)  (succ≺succ m≺n)  = succ≺succ (trans k≺m m≺n)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Exclusiveness/no confusion.

<<decide-helper>>
\begin{code}
not-equal : ∀ {m n} -> m ≺ n -> ¬ m ≡ n
not-equal (zero≺succ)     = λ ()
not-equal (succ≺succ m≺n) = λ where
  reflexive → not-equal m≺n reflexive

not-equal′ : ∀ {m n} -> m ≺ n -> ¬ m ≡ n
not-equal′ p reflexive = ≺-irreflexive p

not-greater : ∀ {m n} -> m ≺ n -> ¬ m ≻ n
not-greater a≺b b≺a = ≺-irreflexive (≺-transitive a≺b b≺a)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Simple variant for the lectures.

\begin{code}
module Private where
\end{code}

<<Maybe>>
\begin{code}
  data Maybe (Property : Set) : Set where
    no  :              Maybe Property
    yes : Property ->  Maybe Property
\end{code}

<<decide-Maybe>>
\begin{code}
  _≺?_ : ∀ m n -> Maybe (m ≺ n)
  m       ≺? zero    = no
  zero    ≺? succ n  = yes zero≺succ
  succ m  ≺? succ n with m ≺? n
  ... | no           = no
  ... | yes m≺n      = yes (succ≺succ m≺n)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Decidability.

Inequality is decidable i.e. given two natural numbers we can either
construct a proof of m ≺ n or a proof of ¬ m ≺ n.

<<decide>>
\begin{code}
infix 6 _≺?_
_≺?_ : ∀ m n -> Decide (m ≺ n)
m ≺? n with compare m n
... | LT {ev = m≺n}       = yes m≺n
... | EQ {ev = reflexive} = no ≺-irreflexive
... | GT {ev = m≻n}       = no (λ m≺n → ≺-irreflexive (≺-transitive m≺n m≻n))
\end{code}

m ≺? n with compare m n
... | LT {ev = m≺n}       = yes m≺n
... | EQ {ev = m≡n} = no  (flip not-equal m≡n)
... | GT {ev = m≻n} = no  (flip not-greater m≻n)

\begin{code}
≺-then-≼ : ∀ {a b c} → a ≺ b → b ≼ c → a ≺ c
≺-then-≼ a≺b (inl reflexive) = a≺b
≺-then-≼ a≺b (inr b≺c)       = ≺-transitive a≺b b≺c

infix 6 _≼?_
_≼?_ : ∀ m n -> Decide (m ≼ n)
m ≼? n with compare m n
... | LT {ev = m≺n} = yes (inr m≺n)
... | EQ {ev = m≡n} = yes (inl m≡n)
... | GT {ev = m≻n} = no (λ m≼n → ≺-irreflexive (≺-then-≼ m≻n m≼n))
\end{code}

no (λ m≺n → ≺-irreflexive (≺-transitive m≺n m≻n))

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Properties of the non-strict order ≼.

\begin{code}
≼-reflexive : n ≼ n
≼-reflexive = inl reflexive

≼-transitive : k ≼ m → m ≼ n → k ≼ n
≼-transitive (inl reflexive) (k≼n)           = k≼n
≼-transitive (inr k≺m)       (inl reflexive) = inr k≺m
≼-transitive (inr k≺m)       (inr m≺n)       = inr (≺-transitive k≺m m≺n)

≼-antisymmetric : m ≼ n → n ≼ m → m ≡ n
≼-antisymmetric (inl reflexive) n≼m             = reflexive
≼-antisymmetric (inr m≺n)       (inl reflexive) = reflexive
≼-antisymmetric (inr m≺n)       (inr n≺m)       = ex-falso-quodlibet (not-greater m≺n n≺m)
\end{code}

|succ≼succ| shows that the successor function is monotone.

\begin{code}
zero≼n : ∀ {n} -> zero ≼ n
zero≼n {zero}   = inl reflexive
zero≼n {succ n} = inr zero≺succ
 
succ≼succ : m ≼ n -> succ m ≼ succ n
succ≼succ (inl reflexive) = inl reflexive
succ≼succ (inr y)         = inr (succ≺succ y)

succ-order-reflecting : succ m ≼ succ n -> m ≼ n
succ-order-reflecting (inl reflexive)     = inl reflexive
succ-order-reflecting (inr (succ≺succ y)) = inr y
\end{code}

begin{code}
zero-bottom : ∀ {n} -> zero ≼ n
zero-bottom {zero} = inl reflexive
zero-bottom {succ n} = inr zero≺succ
end{code}

Decidability.

\begin{code}
decide-≼ : ∀ m n -> Decide (m ≼ n)
decide-≼ m n with compare m n
... | LT {ev = m≺n} = yes (inr m≺n)
... | EQ {ev = m≡n} = yes (inl m≡n)
... | GT {ev = m≻n} = no λ where
  (inl m≡n) → not-equal m≻n (symmetric m≡n) -- OR: ≺-irreflexive m≻n
  (inr m≺n) → not-greater m≺n m≻n

-- ad-hoc definition from scratch
_≼!_ : ∀ m n -> Decide (m ≼ n)
zero   ≼! zero   = yes (inl reflexive)
zero   ≼! succ j = yes (inr zero≺succ)
succ i ≼! zero   = no λ where
                        (inl ())
                        (inr ())
succ i ≼! succ j with i ≼! j
... | no refute           = no λ where
                                 (inl reflexive)       → refute (inl reflexive)
                                 (inr (succ≺succ i≺j)) → refute (inr i≺j)
... | yes (inl reflexive) = yes (inl reflexive)
... | yes (inr i≺j)       = yes (inr (succ≺succ i≺j))
\end{code}

private
  infix 6 _≼?_
  _≼?_ : ∀ m n → m ≼ n ⊎ n ≼ m
  zero   ≼? zero   = inl (inl reflexive)
  zero   ≼? succ n = inl (inr zero≺succ)
  succ m ≼? zero   = inr (inr zero≺succ)
  succ m ≼? succ n with m ≼? n
  ... | inl (inl reflexive) = inl (inl reflexive)
  ... | inl (inr m≺n)       = inl (inr (succ≺succ m≺n))
  ... | inr (inl reflexive) = inl (inl reflexive)
  ... | inr (inr n≺m)       = inr (inr (succ≺succ n≺m))

\begin{code}
order : ∀ m n → m ≼ n ⊎ n ≼ m
order zero zero = inl (inl reflexive)
order zero (succ n) = inl (inr zero≺succ)
order (succ m) zero = inr (inr zero≺succ)
order (succ m) (succ n) with order m n
... | inl m≼n = inl (succ≼succ m≼n)
... | inr n≼m = inr (succ≼succ n≼m)
\end{code}
order = _≼?_

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Relating strict and non-strict ordering (reverse implication of
not-equal and not-greater).
 
\begin{code}
not-leq ¬≼-is-≻ : ∀ {m n} -> ¬ m ≼ n -> m ≻ n -- n ≺ m
not-leq {zero}   {zero}   p = ex-falso-quodlibet (p (inl reflexive))
not-leq {zero}   {succ n} p = ex-falso-quodlibet (p (inr zero≺succ))
not-leq {succ m} {zero}   p = zero≺succ
not-leq {succ m} {succ n} p = succ≺succ (not-leq λ where
                                                   (inl reflexive) → p (inl reflexive)
                                                   (inr m≺n)       → p (inr (succ≺succ m≺n)))
¬≼-is-≻ = not-leq

not-less : ∀ {m n} -> ¬ m ≼ n -> n ≼ m  -- TODO: rename
not-less ¬m≼n = inr (not-leq ¬m≼n)

¬≺-is-≽ : ∀ {m n} -> ¬ m ≺ n -> m ≽ n
¬≺-is-≽ {zero} {zero} p = inl reflexive
¬≺-is-≽ {zero} {succ n} p = ex-falso-quodlibet (p zero≺succ)
¬≺-is-≽ {succ m} {zero} p = inr zero≺succ
¬≺-is-≽ {succ m} {succ n} p with ¬≺-is-≽ {m} {n} (λ x → p (succ≺succ x))
... | inl reflexive = inl reflexive
... | inr y = inr (succ≺succ y)

≼-is-¬≻ : ∀ {m n} -> m ≼ n -> ¬ m ≻ n
≼-is-¬≻ {succ m} {zero} (inl ()) m≻n
≼-is-¬≻ {succ m} {zero} (inr ()) m≻n
≼-is-¬≻ {succ m} {succ .m} (inl reflexive) (succ≺succ m≻n) = ≺-irreflexive m≻n
≼-is-¬≻ {succ m} {succ n} (inr (succ≺succ y)) (succ≺succ m≻n) = ≼-is-¬≻ (inr y) m≻n

≺-is-¬≽ : ∀ {m n} -> m ≺ n -> ¬ m ≽ n
≺-is-¬≽ p q = ≼-is-¬≻ q p
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Impossible combinations.

\begin{code}
impossible₁ : ∀ {A : Set} -> ∀ {n} → n ≺ n → A
impossible₁ n≺n = ex-falso-quodlibet (≺-irreflexive n≺n)

impossible : ∀ {A : Set} -> ∀ {m n} → m ≺ n → n ≺ m -> A
impossible p q = ex-falso-quodlibet (not-greater q p)
--  impossible a≺b b≺a = ex-falso-quodlibet (≺-irreflexive (≺-transitive a≺b b≺a))

impossible₂ : ∀ {A : Set} -> ∀ {m n} → m ≺ n → n ≼ m -> A
impossible₂ m≺n (inl reflexive) = ex-falso-quodlibet (≺-irreflexive m≺n)
impossible₂ m≺n (inr n≺m)       = ex-falso-quodlibet (not-greater m≺n n≺m)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Relating strict and non-strict ordering via arithmetic. TODO: clean up.

\begin{code}
non-strict→ : ∀ {m n} → m ≼ n → m ≺ succ n
non-strict→ (inl reflexive) = n≺succ-n
non-strict→ (inr m≺n)       = ≺-transitive m≺n n≺succ-n

non-strict← : ∀ {m n} → m ≺ succ n → m ≼ n
non-strict← {n = zero}   (zero≺succ) = inl reflexive
non-strict← {n = succ n} (zero≺succ) = inr zero≺succ
non-strict← {n = succ n} (succ≺succ m≺n) with non-strict← m≺n
... | inl reflexive = inl reflexive
... | inr m≺n       = inr (succ≺succ m≺n)

succ-≼-is-≺ : ∀ {a b} -> succ a ≼ b -> a ≺ b
succ-≼-is-≺ (inl reflexive) = n≺succ-n
succ-≼-is-≺ (inr y) = ≺-transitive n≺succ-n y

≼-is-≺-succ : ∀ {a b} -> a ≼ b -> a ≺ succ b  -- same as non-strict→
≼-is-≺-succ (inl reflexive) = n≺succ-n
≼-is-≺-succ (inr y) = ≺-transitive y n≺succ-n
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Relating Boolean-valued and Set-valued comparisons.

\begin{code}
open import VFP.Boolean

<-to : ∀ {a b} -> a ≺ b -> So (a < b)
<-to (zero≺succ)   = yes
<-to (succ≺succ p) = <-to p

<-fro : ∀ {a b} -> So (a < b) -> a ≺ b
<-fro {zero}   {succ b} yes = zero≺succ
<-fro {succ a} {succ b} p   = succ≺succ (<-fro p)

≥-fro : ∀ {a b} -> So (a ≥ b) -> a ≽ b
≥-fro {zero} {zero} yes = inl reflexive
≥-fro {succ a} {zero} yes = inr zero≺succ
≥-fro {succ a} {succ b} p = succ≼succ (≥-fro p)
\end{code}

-------------------------------------------------------------------------------

Properties of arithmetic operations (monotonicity, distributivity, etc).

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Properties of minimum ↓.

\begin{code}
↓-idempotent : ∀ n → n ↓ n ≡ n
↓-idempotent n with compare n n
… | LT {ev = x} = ex-falso-quodlibet (≺-irreflexive x)
--… | EQ {ev = reflexive} = reflexive
… | EQ {ev = ev} = reflexive
… | GT {ev = x} = ex-falso-quodlibet (≺-irreflexive x)
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Monotonicity: addition is strictly monotone in both arguments.

\begin{code}
private
  help : ∀ n k → k ≺ succ (n + k)
  help zero k = n≺succ-n
  help (succ n) k = ≺-transitive n≺succ-n (succ≺succ (help n k))

≺-monotone-left-+ : ∀ {k m n} -> m ≺ n -> m + k ≺ n + k
≺-monotone-left-+ zero≺succ = help _ _
≺-monotone-left-+ (succ≺succ p) = succ≺succ (≺-monotone-left-+ p)

≺-monotone-right-+ : ∀ {k m n} -> m ≺ n -> k + m ≺ k + n
≺-monotone-right-+ {k = zero}   p = p
≺-monotone-right-+ {k = succ k} p = succ≺succ (≺-monotone-right-+ p)

≼-monotone-left-+ : ∀ {k m n} -> m ≼ n -> m + k ≼ n + k
≼-monotone-left-+ (inl reflexive) = inl reflexive
≼-monotone-left-+ (inr y) = inr (≺-monotone-left-+ y)

≼-monotone-right-+ : ∀ {k m n} -> m ≼ n -> k + m ≼ k + n
≼-monotone-right-+ (inl reflexive) = ≼-reflexive
≼-monotone-right-+ (inr y) = inr (≺-monotone-right-+ y)
\end{code}

\begin{code}
pred-monotonic : ∀ {m n : ℕ} → m ≼ n → pred m ≼ pred n
pred-monotonic (inl reflexive)       = inl reflexive
pred-monotonic (inr zero≺succ)       = zero≼n
pred-monotonic (inr (succ≺succ m≺n)) = inr m≺n

≼-monotonic-2* : ∀ {a b} → a ≼ b → 2* a ≼ 2* b
≼-monotonic-2* {zero} {b} p = zero≼n
≼-monotonic-2* {succ a} {zero} (inl ())
≼-monotonic-2* {succ a} {zero} (inr ())
≼-monotonic-2* {succ a} {succ .a} (inl reflexive) = inl reflexive
≼-monotonic-2* {succ a} {succ b} (inr (succ≺succ y)) = succ≼succ (succ≼succ (≼-monotonic-2* (inr y)))

≺-monotonic-2* : ∀ {a b} → a ≺ b → 2* a ≺ 2* b
≺-monotonic-2* {zero} {succ b} p = zero≺succ
≺-monotonic-2* {succ a} {succ b} (succ≺succ p) = succ≺succ (succ≺succ (≺-monotonic-2* p))

⌊/2⌋-monotonic : ∀ {m n : ℕ} → m ≼ n → ⌊ m /2⌋ ≼ ⌊ n /2⌋
⌊/2⌋-monotonic {zero} {n} p = zero≼n
⌊/2⌋-monotonic {succ m} {zero} (inl ())
⌊/2⌋-monotonic {succ m} {zero} (inr ())
⌊/2⌋-monotonic {succ zero} {succ zero} (inl reflexive) = inl reflexive
⌊/2⌋-monotonic {succ zero} {succ zero} (inr (succ≺succ ()))

⌊/2⌋-monotonic {succ zero} {succ (succ n)} p = inr zero≺succ
⌊/2⌋-monotonic {succ (succ m)} {succ zero} (inr (succ≺succ ()))
⌊/2⌋-monotonic {succ (succ m)} {succ (succ .m)} (inl reflexive) = succ≼succ (⌊/2⌋-monotonic (inl reflexive))
⌊/2⌋-monotonic {succ (succ m)} {succ (succ n)} (inr (succ≺succ (succ≺succ y))) = succ≼succ (⌊/2⌋-monotonic (inr y))
\end{code}

\begin{code}
private
  base : ∀ n → 0 ≺ 2^ n
  base zero = zero≺succ
  base (succ n) = ≺-monotonic-2* (base n)

  base′ : ∀ n → 1 ≼ 2^ n
  base′ n = non-strict← (succ≺succ (base n))

≼-monotonic-2^ : ∀ {a b} → a ≺ b → 2^ a ≺ 2^ b
≼-monotonic-2^ {zero} {succ b} zero≺succ = ≺-then-≼ (succ≺succ zero≺succ) (≼-monotonic-2* (base′ b))
≼-monotonic-2^ {succ a} {succ b} (succ≺succ p) = ≺-monotonic-2* (≼-monotonic-2^ p)
\end{code}

Feijen proof of base case (proof format cannot be used because of
cyclic module dependencies):

    proof
      1
    ≺⟨ succ≺succ zero≺succ ⟩
      2
    ≼⟨ ≼-monotonic-2* (base′ b) ⟩
     2* (2^ b)
    ∎
    where open import VFP.Natural.Order.Reasoning

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Properties of monus ∸.

\begin{code}
∸-positive : ∀ {m n} -> m ≺ n -> zero ≺ n ∸ m
∸-positive zero≺succ = zero≺succ
∸-positive (succ≺succ p) = ∸-positive p

∸-pseudo-monotone : ∀ {a b n} -> n ≺ b -> a ≺ b -> a ∸ n ≺ b ∸ n
∸-pseudo-monotone (zero≺succ) (zero≺succ) = zero≺succ
∸-pseudo-monotone (zero≺succ) (succ≺succ q) = succ≺succ q
∸-pseudo-monotone (succ≺succ p) (zero≺succ) = ∸-positive p
∸-pseudo-monotone (succ≺succ p) (succ≺succ q) = ∸-pseudo-monotone p q

m∸n≼m : ∀ {m n} -> m ∸ n ≼ m
m∸n≼m {zero}   {zero}   = inl reflexive
m∸n≼m {zero}   {succ n} = m∸n≼m {zero} {n}
m∸n≼m {succ m} {zero}   = inl reflexive
m∸n≼m {succ m} {succ n} = (≼-transitive (m∸n≼m) (inr n≺succ-n))

∸-pseudo-antitone : ∀ {a b n} -> b ≺ n -> b ≺ a -> n ∸ a ≺ n ∸ b
∸-pseudo-antitone zero≺succ zero≺succ = ≼-is-≺-succ m∸n≼m
∸-pseudo-antitone (succ≺succ p) (succ≺succ q) = ∸-pseudo-antitone p q
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
↓-monotone : ∀ {a₁ a₂} → a₁ ≼ a₂ → ∀ {b₁ b₂} → b₁ ≼ b₂ → a₁ ↓ b₁ ≼ a₂ ↓ b₂
↓-monotone {a₁ = a₁} {a₂ = a₂} p {b₁ = b₁} {b₂ = b₂} q with compare a₁ b₁ | compare a₂ b₂
… | LT {ev = ev₁} | LT {ev = ev₂} = p
… | LT {ev = ev₁} | EQ {ev = reflexive} = p
… | LT {ev = ev₁} | GT = ≼-transitive (inr ev₁) q
… | EQ {ev = reflexive} | LT = p
… | EQ {ev = reflexive} | EQ = p
… | EQ {ev = reflexive} | GT = q
… | GT {ev = ev₁} | LT {ev = ev₂} = ≼-transitive (inr ev₁) p
… | GT {ev = ev₁} | EQ {ev = reflexive} = q
… | GT {ev = ev₁} | GT {ev = ev₂} = q
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Distributivity.

\begin{code}
succ-distributes-↓ : ∀ {a b} → succ a ↓ succ b ≡ succ (a ↓ b)
succ-distributes-↓ {a} {b} with compare a b
… | LT = reflexive
… | EQ {ev = reflexive} = reflexive
… | GT = reflexive

+-distributes-↓ : ∀ n {a b} → (n + a) ↓ (n + b) ≡ n + (a ↓ b)
+-distributes-↓ zero = reflexive
+-distributes-↓ (succ n) = transitive succ-distributes-↓ (congruent succ (+-distributes-↓ n))

2*-distributes-↓ : ∀ a b → (2* a) ↓ (2* b) ≡ 2* (a ↓ b)
2*-distributes-↓ zero zero = reflexive
2*-distributes-↓ zero (succ b) = reflexive
2*-distributes-↓ (succ a) zero = reflexive
2*-distributes-↓ (succ a) (succ b) = transitive succ-distributes-↓ (transitive (congruent succ succ-distributes-↓) (transitive (congruent (λ hole → 2 + hole) (2*-distributes-↓ a b)) (congruent 2*_ (symmetric succ-distributes-↓))))

2^-distributes-↓ : ∀ a b → 2^ (a ↓ b) ≡ (2^ a) ↓ (2^ b)
2^-distributes-↓ a b with compare a b
2^-distributes-↓ a b | LT {ev = ev} with compare (2^ a) (2^ b)
… | LT = reflexive
… | EQ = reflexive
… | GT {ev = ev′} = impossible (≼-monotonic-2^ ev) ev′
2^-distributes-↓ a b | EQ {ev = reflexive} = symmetric (↓-idempotent (2^ a))
2^-distributes-↓ a b | GT {ev = ev} with compare (2^ a) (2^ b)
… | LT {ev = ev′} = impossible ev′ (≼-monotonic-2^ ev)
… | EQ {ev = ev} = symmetric ev
… | GT = reflexive
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Properties of natural division by 2.
TODO: clean up.

\begin{code}
⌊n/2⌋≺succ-n : ∀ n → ⌊ n /2⌋ ≺ succ n
⌊n/2⌋≺succ-n zero = zero≺succ
⌊n/2⌋≺succ-n (succ zero) = zero≺succ
⌊n/2⌋≺succ-n (succ (succ n)) = succ≺succ (≺-transitive (⌊n/2⌋≺succ-n n) n≺succ-n)

⌊succ-n/2⌋≺succ-n : ∀ n → ⌊ succ n /2⌋ ≺ succ n
⌊succ-n/2⌋≺succ-n zero = zero≺succ
⌊succ-n/2⌋≺succ-n (succ n) = succ≺succ (⌊n/2⌋≺succ-n n)

⌈n/2⌉≼n : ∀ n → ⌈ n /2⌉ ≼ n
⌈n/2⌉≼n zero            = inl reflexive
⌈n/2⌉≼n (succ zero)     = inl reflexive
⌈n/2⌉≼n (succ (succ n)) = succ≼succ (≼-transitive (⌈n/2⌉≼n n) (inr n≺succ-n))

⌊-/2⌋-decreasing : ∀ n -> ⌊ n /2⌋ ≼ n
⌊-/2⌋-decreasing (zero) = zero≼n
⌊-/2⌋-decreasing (succ zero) = zero≼n
⌊-/2⌋-decreasing (succ (succ n)) = succ≼succ (≼-transitive (⌊-/2⌋-decreasing n) (inr n≺succ-n))

⌈-/2⌉-decreasing : ∀ n -> ⌈ n /2⌉ ≼ n  -- ⌊ succ n /2⌋ ≡ ⌈ n /2⌉
⌈-/2⌉-decreasing zero = zero≼n
⌈-/2⌉-decreasing (succ n) = succ≼succ (⌊-/2⌋-decreasing n)
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₁|.

\begin{code}
≺₁-transitive : k ≺₁ m → m ≺₁ n → k ≺₁ n
≺₁-transitive p (one)    = succ p
≺₁-transitive p (succ q) = succ (≺₁-transitive p q)
\end{code}

<<zero-bottom-1>>
\begin{code}
zero-bottom₁ : ∀ {n} -> zero ≺₁ succ n
zero-bottom₁ {zero}   = one
zero-bottom₁ {succ n} = succ (zero-bottom₁ {n})
\end{code}

<<succ-monotone-1>>
\begin{code}
succ-monotone₁ : ∀ {m n} -> m ≺₁ n → succ m ≺₁ succ n   -- difference is unchanged
succ-monotone₁ (one)       = one
succ-monotone₁ (succ m≺₁n) = succ (succ-monotone₁ m≺₁n)
\end{code}

\begin{code}
-- order embedding₁: m ≺₁ n  <->  succ m ≺₁ succ n
embedding₁ : ∀ {m n} -> succ m ≺₁ succ n → m ≺₁ n   -- ditto
embedding₁ {m} {succ m} (one)    = one
embedding₁ {m} {succ n} (succ p) = succ (embedding₁ p)
\end{code}

This version is equivalent to |≺|.

<<to-1>>
\begin{code}
to-≺₁ : ∀ {m n} -> m ≺ n -> m ≺₁ n
to-≺₁ (zero≺succ)     = zero-bottom₁
to-≺₁ (succ≺succ m≺n) = succ-monotone₁ (to-≺₁ m≺n)
\end{code}

<<fro-1>>
\begin{code}
fro-≺₁ : ∀ {m n} -> m ≺₁ n -> m ≺ n
fro-≺₁ (one)    = n≺succ-n
fro-≺₁ (succ p) = ≺-transitive (fro-≺₁ p) n≺succ-n
\end{code}

\begin{code}
not-equal₁ : ∀ m n -> m ≺₁ n -> ¬ m ≡ n
not-equal₁ zero (succ m) (one) = λ ()
not-equal₁ zero (succ n) (succ p) = λ ()
not-equal₁ (succ m) (succ n) p = λ where
  reflexive → not-equal₁ m n (embedding₁ p) reflexive

≺₁-irreflexive : ∀ n -> ¬ n ≺₁ n
≺₁-irreflexive n p = not-equal₁ n n p reflexive
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₂|.

begin{code}
≺₂-transitive : k ≺₂ m → m ≺₂ n → k ≺₂ n
≺₂-transitive (one)    q = succ q
≺₂-transitive (succ p) q = succ (≺₂-transitive p q)

zero-bottom₂ : ∀ {n} -> zero ≺₂ succ n
zero-bottom₂ {zero}   = one
zero-bottom₂ {succ n} = ≺₂-transitive zero-bottom₂ one

succ-monotone₂ : ∀ {m n} -> m ≺₂ n → succ m ≺₂ succ n   -- difference is unchanged
succ-monotone₂ (one)    = one
succ-monotone₂ (succ p) = succ (succ-monotone₂ p)
end{code}

This version is equivalent to |≺|.

begin{code}
to-≺₂ : ∀ {m n} -> m ≺ n -> m ≺₂ n
to-≺₂ (zero≺succ)     = zero-bottom₂
to-≺₂ (succ≺succ m≺n) = succ-monotone₂ (to-≺₂ m≺n)

fro-≺₂ : ∀ {m n} -> m ≺₂ n -> m ≺ n
fro-≺₂ (one)    = n≺succ-n
fro-≺₂ (succ p) = ≺-transitive n≺succ-n (fro-≺₂ p)
end{code}

Properties of |≼₂|.

\begin{code}
≼₂-transitive : k ≼₂ m → m ≼₂ n → k ≼₂ n
≼₂-transitive (zero)   q = q
≼₂-transitive (succ p) q = succ (≼₂-transitive p q)

n≼₂succ-n : ∀ {n} -> n ≼₂ succ n
n≼₂succ-n = succ zero

zero≼₂n : ∀ {n} -> zero ≼₂ n
zero≼₂n {zero}   = zero
zero≼₂n {succ n} = ≼₂-transitive (zero≼₂n {n}) (succ zero)

succ-monotone₂ : ∀ {m n} -> m ≼₂ n → succ m ≼₂ succ n   -- difference is unchanged
succ-monotone₂ (zero)   = zero
succ-monotone₂ (succ p) = succ (succ-monotone₂ p)
\end{code}

\begin{code}
to-≼₂ : ∀ {m n} -> m ≼ n -> m ≼₂ n
to-≼₂ (inl reflexive) = zero
to-≼₂ (inr zero≺succ) = zero≼₂n
to-≼₂ (inr (succ≺succ y)) = succ-monotone₂ (to-≼₂ (inr y))

fro-≼₂ : ∀ {m n} -> m ≼₂ n -> m ≼ n
fro-≼₂ (zero)   = ≼-reflexive
fro-≼₂ (succ p) = ≼-transitive (inr n≺succ-n) (fro-≼₂ p)

to-≺ : ∀ {m n} → succ m ≼₂ n → m ≺ n
to-≺ p with fro-≼₂ p
... | inl reflexive = n≺succ-n
... | inr y = ≺-transitive n≺succ-n y
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₃|.

\begin{code}
zero-bottom₃ : ∀ {n} -> zero ≺₃ succ n
zero-bottom₃ {zero}   = one
zero-bottom₃ {succ n} = ≺₃-transitive zero-bottom₃ one

succ-monotone₃ : ∀ {m n} -> m ≺₃ n → succ m ≺₃ succ n   -- difference is unchanged
succ-monotone₃ (one)               = one
succ-monotone₃ (≺₃-transitive p q) = ≺₃-transitive (succ-monotone₃ p) (succ-monotone₃ q)
\end{code}

This version is equivalent to |≺|.

\begin{code}
to-≺₃ : ∀ {m n} -> m ≺ n -> m ≺₃ n
to-≺₃ (zero≺succ)     = zero-bottom₃
to-≺₃ (succ≺succ m≺n) = succ-monotone₃ (to-≺₃ m≺n)

fro-≺₃ : ∀ {m n} -> m ≺₃ n -> m ≺ n
fro-≺₃ (one)                = n≺succ-n
fro-≺₃ (≺₃-transitive p p₁) = ≺-transitive (fro-≺₃ p) (fro-≺₃ p₁)
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₄|, the variant that relates ordering to addition
using a Σ-type.

\begin{code}
open import VFP.Natural.Properties
open import VFP.Equality.Reasoning

≺₄-transitive : ∀ {k m n} -> k ≺₄ m → m ≺₄ n → k ≺₄ n
≺₄-transitive {k = k} (Δ₁ , reflexive) (Δ₂ , reflexive) = succ (Δ₂ + Δ₁) , congruent succ (
    succ ((Δ₂ + Δ₁) + k)
  ≡⟨ +-associative (succ Δ₂) ⟩ 
    succ (Δ₂ + (Δ₁ + k))
  ≡⟨ symmetric (shift Δ₂ (Δ₁ + k)) ⟩ 
    Δ₂ + succ (Δ₁ + k)
  ∎ )
\end{code}

This version is equivalent to |≺₁|.

<<to-fro-4>>
\begin{code}
to-≺₄ : ∀ {m n} -> m ≺₁ n -> m ≺₄ n
to-≺₄ (one)    = 0 , reflexive
to-≺₄ (succ p) with to-≺₄ p
... | Δ , m≺₁n = succ Δ , congruent succ m≺₁n

fro-≺₄ : ∀ {m n} -> m ≺₄ n -> m ≺₁ n
fro-≺₄ (zero   , reflexive) = one
fro-≺₄ (succ Δ , reflexive) = succ (fro-≺₄ (Δ , reflexive))
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₅|.

\begin{code}
≺₅-transitive : ∀ {k m n} -> k ≺₅ m → m ≺₅ n → k ≺₅ n
≺₅-transitive {k = k} (Δ₁ , reflexive) (Δ₂ , reflexive) = succ (Δ₁ + Δ₂) , (
    succ (k + succ (Δ₁ + Δ₂))
  ≡⟨ congruent succ (shift k (Δ₁ + Δ₂)) ⟩ 
    succ (succ (k + (Δ₁ + Δ₂)))
  ≡⟨ congruent (λ hole → succ (succ hole)) (symmetric (+-associative k)) ⟩ 
    succ (succ ((k + Δ₁) + Δ₂))
  ∎ )
\end{code}

TODO: equivalence to |≺₁|.
