Feijen proof format for strict relations - this is more difficult as
the end-of-proof symbol ∎ relies on reflexibility. 

    proof
      ?
    ≡⟨ ? ⟩
    ≺⟨ ? ⟩
    ≼⟨ ? ⟩
     ?
    ∎
    where open VFP.Natural.Order.Reasoning

The format is quite flexible: it's an equality proof (x ≡ y) if only
≡-steps are used; it establishes a strict order (x ≺ y) if there is at
least one ≺-step; othwerwise it shows a non-strict order (x ≼ y).

See also
Relation.Binary.Reasoning.Preorder
https://plfa.github.io/19.08/Equality/
https://gist.github.com/UlfNorell/102f8e59744de8a69f94b39f09f70c1d

\begin{code}
{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --safe #-}

module VFP.Natural.Order.Reasoning where

open import VFP.Natural.Order.Properties public
open import VFP.Logic

private
  variable
    m n : ℕ
  data Order : ℕ -> ℕ -> Set where
    LT : m ≺ n -> Order m n
    LE : m ≼ n -> Order m n
    EQ : m ≡ n -> Order m n

-- an anonymous module so that ⟦_⟧ and proof share the same arguments: i and j  
module _ {i j : ℕ} where
  ⟦_⟧ : Order i j -> Set
  ⟦ LT  _ ⟧ = i ≺ j
  ⟦ LE _ ⟧ = i ≼ j
  ⟦ EQ  _ ⟧ = i ≡ j

  infix 0 proof_

  proof_ : (p : Order i j) → ⟦ p ⟧
  proof (LT p) = p
  proof (LE p) = p
  proof (EQ p) = p

infixr 1 ≺-link ≼-link ≡-link 
infix  2 _∎
syntax ≺-link x q p = x ≺⟨ p ⟩ q
syntax ≼-link x q p = x ≼⟨ p ⟩ q
syntax ≡-link x q p = x ≡⟨ p ⟩ q
-- order of last two arguments is swapped so that p's type can be inferred easily

-- the result type is x ≡ y if only ≡-links are used; it is x ≺ y if there is
-- at least one ≺-link; othwerwise it is x ≼ y

infix 0 case_of_
case_of_ : {A : Set} {B : Set} → A → (A → B) → B
case x of f = f x

≺-link : (x : ℕ) {y z : ℕ} → Order y z → x ≺ y → Order x z
≺-link x (LT y≺z) x≺y  = LT (≺-transitive x≺y y≺z)
≺-link x (LE y≼z) x≺y  = LT (case y≼z of λ { (inl reflexive) -> x≺y ; (inr y≺z) -> ≺-transitive x≺y y≺z})
≺-link x (EQ y≡z) x≺y  = LT (case y≡z of λ where reflexive → x≺y)

≼-link : (x : ℕ) {y z : ℕ} → Order y z → x ≼ y → Order x z
≼-link x (LT y≺z) x≼y  = LT (case x≼y of λ { (inl reflexive) -> y≺z ; (inr x≺y) -> ≺-transitive x≺y y≺z})
≼-link x (LE y≼z) x≼y  = LE (≼-transitive x≼y y≼z)
≼-link x (EQ y≡z) x≼y  = LE (case y≡z of λ where reflexive → x≼y)

≡-link : (x : ℕ) {y z : ℕ} → Order y z → x ≡ y → Order x z
≡-link x (LT y≺z) x≡y  = LT (case x≡y of λ where reflexive -> y≺z)
≡-link x (LE y≼z) x≡y  = LE (case x≡y of λ where reflexive -> y≼z)
≡-link x (EQ y≡z) x≡y  = EQ (case x≡y of λ where reflexive -> y≡z)
-- |case| on the RHS instead of PM on the left: delayed
-- unification. We need to be able determine the "kind" of Order based
-- on the "kinds" of the arguments (so that |proof| can infer the
-- result type).

-- Ulf Norell says:
-- ^ Note: don't match on the proof here, we need to decide strict vs nonstrict for neutral proofs

_∎ : (x : ℕ) → Order x x
x ∎ = EQ reflexive
\end{code}
