Source file for the paper „Functional Pearl: Type-directed, Modular
Huffman Encoding“, original submission (the paper was accepted with
major revisions). Status of this file: frozen, see Serialise.Indexed
for an updated version.

-------------------------------------------------------------------------------

\begin{code}
module VFP.Serialise where

open import VFP.Equality.Properties public 
open import VFP.Logic public
open import VFP.Maybe using (Maybe ; undefined ; Just ; invoke) public
import VFP.Maybe as Maybe

private variable
  A B : Set

infixr 10 _·_

infixr 1 _●_
_●_ = transitive
\end{code}

-------------------------------------------------------------------------------

Bit strings, isomorphic to List (⊤ ⊎ ⊤).

<<Bits>>
\begin{code}
data Bits : Set where
  []   : Bits
  𝟢 𝟣  : Bits -> Bits
\end{code}

\begin{code}
Eau-de-Cologne = 𝟣 (𝟣 (𝟣 (𝟢 (𝟢 (𝟣 (𝟣 (𝟢 (𝟢 (𝟣 (𝟢 (𝟢 (𝟣 []))))))))))))
\end{code}

\begin{code}
module Natural where
  open import VFP.Natural renaming (zero to z ; succ to s)

  succ : Bits -> Bits
  succ []    = 𝟣 []
  succ (𝟢 n) = 𝟣 n
  succ (𝟣 n) = 𝟢 (succ n)

  from-Natural : Natural -> Bits
  from-Natural (z)   = []
  from-Natural (s n) = succ (from-Natural n)
\end{code}

-------------------------------------------------------------------------------

Simplified version of Haskell's Show and Read classes. Printer: uses
an accumulating argument for efficiency (Hughes' efficient
representation of lists). Deterministic, but partial parser: returns
residual input; combination of Maybe and state monad.

<<Decoder-Encoder>>
\begin{code}
Encoder Decoder : Set -> Set
Encoder A  = A -> Bits -> Bits
Decoder A  = Bits -> Maybe (A × Bits)
\end{code}

<<correct>>
\begin{code}
_↬_ : Encoder A -> Decoder A -> Set
encode ↬ decode = ∀ a -> ∀ {bs} -> decode (encode a bs) ≡ Just (a , bs)
\end{code}

<<Coder>>
\begin{code}
record Coder (A : Set) : Set where
  field
    encode   : Encoder A
    decode   : Decoder A
    correct  : encode ↬ decode
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Combinators for encoding aka printing: |ε| is a synonym for the
identity, |·| for function composition.

<<encoding-combinators>>
\begin{code}
ε : Bits -> Bits
ε = λ bs -> bs

_·_ : (Bits -> Bits) -> (Bits -> Bits) -> (Bits -> Bits)
f · g = λ bs -> f (g bs)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Combinators for decoding aka parsing.

<<decoding-combinators>>
\begin{code}
return : A -> Decoder A
return a bs = Maybe.return (a , bs)

_>>=_ : Decoder A -> (A -> Decoder B) -> Decoder B
m >>= k = λ bs -> m bs Maybe.>>= λ (a , bs′) -> k a bs′

case-𝟢:_case-𝟣:_ : Decoder A -> Decoder A -> Decoder A
(case-𝟢: f case-𝟣: g) []      = undefined
(case-𝟢: f case-𝟣: g) (𝟢 bs)  = f bs
(case-𝟢: f case-𝟣: g) (𝟣 bs)  = g bs
\end{code}

Monad laws.

\begin{code}
open import VFP.Function.Equality

private variable
  C : Set
  a : A
  m : Decoder A
  f : A -> Decoder B
  g : B -> Decoder C

left-identity  : return a >>= f ⩧ f a
right-identity : m >>= return ⩧ m
associativity  : m >>= (λ a -> f a >>= g) ⩧ (m >>= f) >>= g

left-identity         bs = reflexive
right-identity        bs = Maybe.right-identity 
associativity {m = m} bs = Maybe.associativity {m = m bs}
\end{code}

-------------------------------------------------------------------------------

A mappping function …

\begin{code}
open import VFP.Bijection.Properties
\end{code}

<<Map-encode-decode>>
\begin{code}
module Coder-Map (iso : A ≅ B) (𝔸 : Coder A) where
  encode : Encoder B
  encode b = Coder.encode 𝔸 (fro iso b)

  decode : Decoder B
  decode =  do a <- Coder.decode 𝔸 ; return (to iso a)
\end{code}

\begin{code}
  open import VFP.Equality.Reasoning
  correct : encode ↬ decode
  correct b {bs = bs} =
      decode (encode b bs)
    ≡⟨ reflexive ⟩
      (do a <- Coder.decode 𝔸 ; return (to iso a)) (Coder.encode 𝔸 (fro iso b) bs)
    ≡⟨ invoke (Coder.correct 𝔸 (fro iso b)) ⟩
      (return (to iso (fro iso b))) bs
    ≡⟨ congruent (λ hole -> Just (hole , bs)) (to∘fro≡id iso b) ⟩ 
      Just (b , bs)
    ∎
\end{code}

<<Map-record>>
\begin{code}
𝕄𝕒𝕡 : (A ≅ B) -> (Coder A -> Coder B)
\end{code}

\begin{code}
𝕄𝕒𝕡 iso 𝔸 = record { Coder-Map iso 𝔸 }
\end{code}

-------------------------------------------------------------------------------

Encoders and decoders for some (standard) types.

\begin{code}
open import VFP.Boolean

module Coder-Boolean where
  encode : Encoder 𝔹
  encode false = 𝟢
  encode true  = 𝟣

  decode : Decoder 𝔹
  decode = case-𝟢: return false case-𝟣: return true

  correct : encode ↬ decode
  correct false = reflexive
  correct true  = reflexive

𝔹𝕠𝕠𝕝𝕖𝕒𝕟 : Coder Boolean
𝔹𝕠𝕠𝕝𝕖𝕒𝕟 = record { Coder-Boolean }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
open import VFP.Natural
\end{code}

<<Natural-encode>>
\begin{code}
module Coder-Natural where
  encode : Encoder Natural
  encode (zero)    = 𝟢
  encode (succ n)  = 𝟣 · encode n
\end{code}

<<Natural-decode>>
\begin{code}
  {-# TERMINATING #-}
  decode : Decoder Natural
  decode =  case-𝟢:  (return zero)
            case-𝟣:  (do n <- decode ; return (succ n))
\end{code}

Agda's termination checker is happy with the following version where
all the combinators are inlined.

\begin{code}
  decode′ : Decoder Natural
  decode′ []     = undefined
  decode′ (𝟢 bs) = Just (zero , bs)
  decode′ (𝟣 bs) = decode′ bs Maybe.>>= λ (n , bs′) -> Just (succ n , bs′)
\end{code}

<<Natural-correct>>
\begin{code}
  correct : encode ↬ decode
  correct (zero)    = reflexive
  correct (succ n)  = invoke (correct n)
\end{code}

<<Natural-record>>
\begin{code}
ℕ𝕒𝕥𝕦𝕣𝕒𝕝 : Coder Natural
ℕ𝕒𝕥𝕦𝕣𝕒𝕝 = record { Coder-Natural }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
open import VFP.Ternary hiding (_𝟘)

module Coder-Ternary where
  encode : Encoder Ternary
  encode (𝟘t)  = 𝟢 · 𝟢
  encode (n 𝟙) = 𝟢 · 𝟣 · encode n
  encode (n 𝟚) = 𝟣 · 𝟢 · encode n
  encode (n 𝟛) = 𝟣 · 𝟣 · encode n

  {-# TERMINATING #-}
  decode : Decoder Ternary
  decode = case-𝟢: (case-𝟢: (return 𝟘t)
                    case-𝟣: (do n <- decode ; return (n 𝟙)))
           case-𝟣: (case-𝟢: (do n <- decode ; return (n 𝟚))
                    case-𝟣: (do n <- decode ; return (n 𝟛)))

  correct : encode ↬ decode
  correct (𝟘t)  = reflexive
  correct (n 𝟙) = invoke (correct n)
  correct (n 𝟚) = invoke (correct n)
  correct (n 𝟛) = invoke (correct n)

𝕋𝕖𝕟𝕒𝕣𝕪 : Coder Ternary
𝕋𝕖𝕟𝕒𝕣𝕪 = record { Coder-Ternary }
\end{code}

In general, n constructors can be encoded with ⌈ lg n ⌉ bits.

Again, inlining the combinators resolves the issue with the
termination checker.

begin{code}
  decode′ : Decoder Ternary
  decode′ []         = undefined
  decode′ (𝟢 [])     = undefined
  decode′ (𝟢 (𝟢 bs)) = Just (𝟘t , bs)
  decode′ (𝟢 (𝟣 bs)) = decode′ bs Maybe.>>= λ (n , bs′) -> Just (n 𝟙 , bs′)
  decode′ (𝟣 [])     = undefined
  decode′ (𝟣 (𝟢 bs)) = decode′ bs Maybe.>>= λ (n , bs′) -> Just (n 𝟚 , bs′)
  decode′ (𝟣 (𝟣 bs)) = decode′ bs Maybe.>>= λ (n , bs′) -> Just (n 𝟛 , bs′)
end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<Bush>>
\begin{code}
data Bush : Set where
  Leaf  : Bush
  Fork  : Bush -> Bush -> Bush
\end{code}
  
<<Coder-Bush>>
\begin{code}
module Coder-Bush where
  encode : Encoder Bush
  encode (Leaf)      = 𝟢
  encode (Fork t u)  = 𝟣 · encode t · encode u

  {-# TERMINATING #-}
  decode : Decoder Bush
  decode =  case-𝟢:  (return Leaf)
            case-𝟣:  (do t <- decode ; u <- decode ; return (Fork t u))

  correct : encode ↬ decode
  correct (Leaf)      = reflexive
  correct (Fork t u)  = invoke (correct t) ● invoke (correct u)

𝔹𝕦𝕤𝕙 : Coder Bush
𝔹𝕦𝕤𝕙 = record { Coder-Bush }
\end{code}

Alas, inlining the combinators doesn't always help. The second call to
decode′ works on the *output* of the first call. A priori there is no
reasong to believe that the output is strictly smaller than the input.
Indeed, this is the principle problem of LL-parsing: it may loop for
left-recursive productions.

begin{code}
  {-# TERMINATING #-}
  decode′ : Decoder Bush
  decode′ []     = undefined
  decode′ (𝟢 bs) = Just (Leaf , bs)
  decode′ (𝟣 bs) = decode′ bs Maybe.>>= λ (t , bs′) -> decode′ bs′ Maybe.>>= λ (u , bs″) -> Just (Fork t u , bs″)
end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Serializing elements of dependent types. Vectors: no bits required!

\begin{code}
open import VFP.Vector
\end{code}

<<Vector-encode-decode>>
\begin{code}
module Coder-Vector {A : Set} (𝔸 : Coder A) where
  encode : {n : Natural} -> Encoder (Vector A n)
  encode ⟨⟩        = ε
  encode (a , as)  = Coder.encode 𝔸 a · encode as

  decode : (n : Natural) -> Decoder (Vector A n)
  decode (zero)    = return ⟨⟩
  decode (succ n)  = do a ← Coder.decode 𝔸 ; as ← decode n ; return (a , as)
\end{code}

\begin{code}
  correct : {n : Natural} -> encode ↬ decode n
  correct ⟨⟩        = reflexive
  correct (a , as)  = invoke (Coder.correct 𝔸 a) ● invoke (correct as)
\end{code}

<<Vector-record>>
\begin{code}
𝕍𝕖𝕔𝕥𝕠𝕣 : {A : Set} -> Coder A -> (n : Natural) -> Coder (Vector A n)
\end{code}

\begin{code}
𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n = record
  { encode  = Coder-Vector.encode  𝔸
  ; decode  = Coder-Vector.decode  𝔸 n
  ; correct = Coder-Vector.correct 𝔸 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<List>>
\begin{code}
open import VFP.List using (List ; [] ; _∷_)
\end{code}

<<List-encode-decode>>
\begin{code}
module Coder-List {A : Set} (𝔸 : Coder A) where
  encode : Encoder (List A)
  encode []        = 𝟢
  encode (a ∷ as)  = 𝟣 · Coder.encode 𝔸 a · encode as

  {-# TERMINATING #-}
  decode : Decoder (List A)
  decode =  case-𝟢:  return []
            case-𝟣:  do a ← Coder.decode 𝔸 ; as ← decode ; return (a ∷ as)
\end{code}

Same problem here: the recursive call is applied to the output of a
previous parser.

\begin{code}
  {-# TERMINATING #-}
  decode′ : Decoder (List A)
  decode′ []     = undefined
  decode′ (𝟢 bs) = Just ([] , bs)
  decode′ (𝟣 bs) = Coder.decode 𝔸 bs Maybe.>>= λ (a , bs′) -> decode′ bs′ Maybe.>>= λ (as , bs″) -> Just (a ∷ as , bs″)
\end{code}

\begin{code}
  correct : encode ↬ decode
  correct []       = reflexive
  correct (a ∷ as) = invoke (Coder.correct 𝔸 a) ● invoke (correct as)
\end{code}

<<List-record>>
\begin{code}
𝕃𝕚𝕤𝕥 : {A : Set} -> Coder A -> Coder (List A)
\end{code}

\begin{code}
𝕃𝕚𝕤𝕥 𝔸 = record { Coder-List 𝔸 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Serializer for dependent sums/products aka Sigma types.

<<Sigma-encode-decode>>
\begin{code}
module Coder-Σ  {A : Set} (𝔸 : Coder A) 
                  {B : A → Set} (𝔹 : (a : A) -> Coder (B a)) where
  encode : Encoder (Σ A B)
  encode (a , b) = Coder.encode 𝔸 a · Coder.encode (𝔹 a) b

  decode : Decoder (Σ A B)
  decode = do a <- Coder.decode 𝔸 ; b <- Coder.decode (𝔹 a) ; return (a , b)
\end{code}

\begin{code}
  correct : encode ↬ decode
  correct (a , b) = invoke (Coder.correct 𝔸 a) ● invoke (Coder.correct (𝔹 a) b)
\end{code}

<<Sigma-record>>
\begin{code}
𝕊𝕚𝕘𝕞𝕒 :  {A : Set} -> Coder A -> 
            {B : A → Set} -> ((a : A) -> Coder (B a)) -> 
              Coder (Σ A B)
\end{code}

\begin{code}
𝕊𝕚𝕘𝕞𝕒 𝔸 𝔹 = record { Coder-Σ 𝔸 𝔹 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<List-prime>>
\begin{code}
List′ : Set -> Set
List′ A = Σ Natural (Vector A)
\end{code}

Ad-hoc encoder and decoder (more efficient version further below).

\begin{code}
module Coder-List′ {A : Set} (𝔸 : Coder A) where
  encode : Encoder (List′ A)
  encode (n , as) = Coder.encode ℕ𝕒𝕥𝕦𝕣𝕒𝕝 n · Coder.encode (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) as

  decode : Decoder (List′ A)
  decode = do n <- Coder.decode ℕ𝕒𝕥𝕦𝕣𝕒𝕝 ; as <- Coder.decode (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) ; return (n , as)

  correct : encode ↬ decode
  correct (n , as) = invoke (Coder.correct ℕ𝕒𝕥𝕦𝕣𝕒𝕝 n) ● invoke (Coder.correct (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) as)

ad-hoc-𝕃𝕚𝕤𝕥′ : {A : Set} -> Coder A -> Coder (List′ A)
ad-hoc-𝕃𝕚𝕤𝕥′ 𝔸 = record { Coder-List′ 𝔸 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
Ternary≅Natural = ≅-symmetric Natural≅Ternary
\end{code}

<<Ternary>>
\begin{code}
ℕ𝕒𝕥𝕦𝕣𝕒𝕝′ : Coder Natural
ℕ𝕒𝕥𝕦𝕣𝕒𝕝′ = 𝕄𝕒𝕡 Ternary≅Natural 𝕋𝕖𝕟𝕒𝕣𝕪
\end{code}

<<List-prime-record>>
\begin{code}
𝕃𝕚𝕤𝕥′ : {A : Set} -> Coder A -> Coder (List′ A)
𝕃𝕚𝕤𝕥′ 𝔸 = 𝕊𝕚𝕘𝕞𝕒 (𝕄𝕒𝕡 Ternary≅Natural 𝕋𝕖𝕟𝕒𝕣𝕪) (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸)
\end{code}

Ctrl-C Ctrl-N: Coder.encode ℕ𝕒𝕥𝕦𝕣𝕒𝕝′ 4711

-------------------------------------------------------------------------------
