Source file for the paper „Functional Pearl: Type-directed, Modular
Huffman Encoding“, original submission (the paper was accepted with
major revisions). Status of this file: frozen, see
Serialize.Indexed.Huffman for an updated version.

Huffman encoding (we assume that the code tree is given to us).

-------------------------------------------------------------------------------

\begin{code}
--{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --without-K #-}

module VFP.Serialise.Huffman where

open import VFP.Character
open import VFP.Equality.Properties
open import VFP.Serialise public

private variable
  A B Alphabet : Set
\end{code}

-------------------------------------------------------------------------------

Serializer for Huffman code trees.

<<Code-Tree>>
\begin{code}
data Code-Tree (Alphabet : Set) : Set where
  [_] : Alphabet -> Code-Tree Alphabet
  _⋏_ : Code-Tree Alphabet -> Code-Tree Alphabet -> Code-Tree Alphabet
\end{code}

<<example>>
\begin{code}
code =  (([ 'h' ] ⋏ [ 't' ]) ⋏ [ 'e' ])
          ⋏ (([ 'i' ] ⋏ [ 'n' ]) ⋏ (([ 'v' ] ⋏ [ 'g' ]) ⋏ [ blank ]))
\end{code}

<<Coder-Code-Tree-encode-decode>>
\begin{code}
module Coder-Code-Tree {A : Set} (𝔸 : Coder A) where
  encode : Encoder (Code-Tree A)
  encode [ a ]    = 𝟢 · Coder.encode 𝔸 a
  encode (t ⋏ u)  = 𝟣 · encode t · encode u

  {-# TERMINATING #-}
  decode : Decoder (Code-Tree A)
  decode =  case-𝟢:  (do a ← Coder.decode 𝔸 ; return [ a ])
            case-𝟣:  (do t ← decode ; u ← decode ; return (t ⋏ u))
\end{code}

\begin{code}
  correct : encode ↬ decode
  correct [ a ]    = invoke (Coder.correct 𝔸 a)
  correct (t ⋏ u)  = invoke (correct t) ● invoke (correct u)
\end{code}

<<Coder-Code-Tree-record>>
\begin{code}
ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 : {A : Set} -> Coder A -> Coder (Code-Tree A)
\end{code}

\begin{code}
ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝔸 = record { Coder-Code-Tree 𝔸 } 
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Classical Huffman encoding.

Existence: evidence/witness is a path into the tree.

<<exists>>
\begin{code}
data exists (P : Alphabet -> Set) : (Code-Tree Alphabet -> Set) where
  here   : ∀ {c}    -> P c         -> exists P [ c ]
  left   : ∀ {t u}  -> exists P t  -> exists P (t ⋏ u)
  right  : ∀ {t u}  -> exists P u  -> exists P (t ⋏ u)
\end{code}

<<i-in-code>>
\begin{code}
i-am-in  :  exists (λ c′ -> 'i' ≡ c′) code
i-am-in  =  right (left (left (here reflexive)))
\end{code}

<<epsilon>>
\begin{code}
_∈_ : (Alphabet : Set) -> Code-Tree Alphabet -> Set
Alphabet ∈ t = Σ Alphabet (λ c -> exists (λ c′ -> c ≡ c′) t)
\end{code}

huffman-encode: driven by the path; huffman-decode: driven by bit
string and the tree.

<<Compress-encode>>
\begin{code}
module Compress where
  encode : ∀ {t : Code-Tree Alphabet} -> Encoder (Alphabet ∈ t)
  encode (c , here   e)  = ε
  encode (c , left   p)  = 𝟢 · encode (c , p)
  encode (c , right  p)  = 𝟣 · encode (c , p)
\end{code}

<<Compress-decode>>
\begin{code}
  decode : ∀ (t : Code-Tree Alphabet) -> Decoder (Alphabet ∈ t)
  decode [ c ]   =  return (c , here reflexive)
  decode (t ⋏ u) =  case-𝟢:  (do (c , p) <- decode t  ; return (c , left   p))
                    case-𝟣:  (do (c , p) <- decode u  ; return (c , right  p))
\end{code}

\begin{code}
  correct : ∀ {t : Code-Tree Alphabet} -> encode ↬ decode t
  correct (c , here reflexive)  = reflexive
  correct (c , left   p)        = invoke (correct (c , p))
  correct (c , right  p)        = invoke (correct (c , p))
\end{code}

<<Compress-record>>
\begin{code}
ℂ𝕠𝕞𝕡𝕣𝕖𝕤𝕤 : ∀ (t : Code-Tree Alphabet) -> Coder (Alphabet ∈ t)
\end{code}

\begin{code}
ℂ𝕠𝕞𝕡𝕣𝕖𝕤𝕤 t = record
  { encode  = Compress.encode
  ; decode  = Compress.decode t
  ; correct = Compress.correct }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Encoding a list of characters paired with its code tree.

<<Huffman>>
\begin{code}
Huffman : Set -> Set
Huffman Alphabet = Σ (Code-Tree Alphabet) (λ t -> List′ (Alphabet ∈ t))
\end{code}

Ad-hoc encoder and decoder.

\begin{code}
module Coder-Huffman {Alphabet : Set} (𝕔𝕙𝕒𝕣 : Coder Alphabet) where
  encode : Encoder (Huffman Alphabet)
  encode (t , as) = Coder.encode (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) t · Coder.encode (𝕃𝕚𝕤𝕥′ (ℂ𝕠𝕞𝕡𝕣𝕖𝕤𝕤 t)) as

  decode : Decoder (Huffman Alphabet)
  decode = do t <- Coder.decode (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) ; as <- Coder.decode (𝕃𝕚𝕤𝕥′ (ℂ𝕠𝕞𝕡𝕣𝕖𝕤𝕤 t)) ; return (t , as)

  correct : encode ↬ decode
  correct (t , as) = invoke (Coder.correct (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) t) ● invoke (Coder.correct (𝕃𝕚𝕤𝕥′ (ℂ𝕠𝕞𝕡𝕣𝕖𝕤𝕤 t)) as)

ℍ𝕦𝕗𝕗𝕞𝕒𝕟′ : {Alphabet : Set} -> Coder Alphabet -> Coder (Huffman Alphabet)
ℍ𝕦𝕗𝕗𝕞𝕒𝕟′ 𝕔𝕙𝕒𝕣 = record { Coder-Huffman 𝕔𝕙𝕒𝕣 }
\end{code}

Or a tad shorter using 𝕊𝕚𝕘𝕞𝕒.

<<Huffman-record>>
\begin{code}
ℍ𝕦𝕗𝕗𝕞𝕒𝕟 : {Alphabet : Set} -> Coder Alphabet -> Coder (Huffman Alphabet)
ℍ𝕦𝕗𝕗𝕞𝕒𝕟 𝕔𝕙𝕒𝕣 = 𝕊𝕚𝕘𝕞𝕒 (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) (λ t -> 𝕃𝕚𝕤𝕥′ (ℂ𝕠𝕞𝕡𝕣𝕖𝕤𝕤 t))
\end{code}

Modularity: lists can be easily replaced by other collection types.

-------------------------------------------------------------------------------
