Source file for the paper „Functional Pearl: Certified, Total
Serialisers With an Application to Huffman Encoding“. It introduces:

●  parsing combinators,
●  printing combinators, and 
●  proof combinators.

A „type class“ bundling the three gadgets is defined in
VFP.Serialise.Indexed.Codec.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Serialise.Indexed where

open import VFP.Bits public
open import VFP.Equality.Reasoning
open import VFP.Logic hiding (_&_)
open import VFP.Maybe using (Maybe ; Nothing ; Just ; undefined)
import VFP.Maybe as Maybe
open import VFP.Natural.Non-Strict-Order hiding (recurse)

private variable
  A B C : Set
  i j k m n : ℕ
\end{code}

-------------------------------------------------------------------------------

Parsing combinators aka decoders.

Parsers form a monad (a combination of Maybe and state). Additionally,
we provide a combinator that performs a case-analysis on the leading bit.

To ensure that decoders are terminating we make heavy use of indexed
types. A decoder takes a bit string of some size below a given bound;
it possibly returns a thing of type A, and the residual input, a
smaller bit string.

<<IDecoder>>
\begin{code}
IDecoder : Set -> (ℕ -> Set)
IDecoder A i = ∀ {j} -> i ≽ j -> Bits j -> Maybe (A × Bits≼ j)
\end{code}

<<Decoder>>
\begin{code}
Decoder : Set -> Set
Decoder A = ∀ {i} -> IDecoder A i
\end{code}

<<tighten>>
\begin{code}
tighten : (i ≽ j) -> (IDecoder A i -> IDecoder A j)
tighten i≽j p = λ j≽k bs -> p (≽-transitive i≽j j≽k) bs
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

<<relax>>
\begin{code}
relax : i ≽ j -> Maybe (A × Bits≼ j) -> Maybe (A × Bits≼ i)
relax i≽j = Maybe.map (λ (a , bs) -> (a , raise i≽j bs))
\end{code}

TODO: use Maybe monad laws below.

\begin{code}
relax-preserves-identity : ∀ (r : Maybe (A × Bits≼ n)) ->
    relax ≽-reflexive r ≡ r
relax-preserves-identity Nothing = reflexive
relax-preserves-identity (Just x) = congruent (λ hole → Just (_ , hole)) (raise-preserves-identity _)

relax-preserves-composition : ∀ (n≽m : n ≽ m) (m≽k : m ≽ k) (r : Maybe (A × Bits≼ k)) ->
    relax (≽-transitive n≽m m≽k) r ≡ relax n≽m (relax m≽k r)
relax-preserves-composition n≽m m≽k Nothing = reflexive
relax-preserves-composition n≽m m≽k (Just (a , bs)) =
    proof
      relax (≽-transitive n≽m m≽k) (Just (a , bs))
    ≡⟨ reflexive ⟩
      Just (a , raise (≽-transitive n≽m m≽k) bs)
    ≡⟨ congruent (λ hole → Just (a , hole)) (raise-preserves-composition n≽m m≽k _) ⟩
      Just (a , raise n≽m (raise m≽k bs))
    ≡⟨ reflexive ⟩ 
      relax n≽m (relax m≽k (Just (a , bs)))
    ∎
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Monadic interface.

<<return>>
\begin{code}
return : A -> IDecoder A i
return a = λ i≽j bs -> Maybe.return (a , bs such-that ≽-reflexive)
\end{code}

NB. The proof i≽j is ignored as we only need to show that the size of
bs is below j, not below i.

<<bind>>
\begin{code}
_>>=_ : IDecoder A i -> (A -> IDecoder B i) -> IDecoder B i
p >>= q = λ i≽j bs ->
  p i≽j bs Maybe.>>= λ (a , bs′ such-that j≽k) ->
    relax j≽k (tighten i≽j (q a) j≽k bs′)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Proof of the monad laws.

\begin{code}
module _ where
  private variable
    a : A
    M : IDecoder A n
    F : A -> IDecoder B n
    G : B -> IDecoder C n

--  left-identity  : return a >>= F ≡ F a
  left-identity : ∀ (n≽m : n ≽ m) (bs : Bits m) ->
      (return a >>= F) n≽m bs ≡ (F a) n≽m bs
  left-identity {a = a} {F = F} n≽m bs =
    proof
      (return a >>= F) n≽m bs
    ≡⟨ reflexive ⟩ 
      (return a n≽m bs  Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (F a (≽-transitive n≽m m≽k) bs′))
    ≡⟨ reflexive ⟩
      relax ≽-reflexive (F a (≽-transitive n≽m ≽-reflexive) bs)
    ≡⟨ relax-preserves-identity _ ⟩ 
      F a (≽-transitive n≽m ≽-reflexive) bs
    ≡⟨ congruent (λ hole → F a hole bs) (≽-right-identity _) ⟩ 
      (F a) n≽m bs
    ∎

--  right-identity : M >>= return ≡ M
  right-identity : ∀ (n≽m : n ≽ m) (bs : Bits m) ->
    (M >>= return) n≽m bs ≡ (M) n≽m bs
  right-identity {M = M} n≽m bs =
    proof
      (M >>= return) n≽m bs
    ≡⟨ reflexive ⟩ 
      (M n≽m bs Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (return a (≽-transitive n≽m m≽k) bs′))
    ≡⟨ Maybe.invoke-right-identity _ (λ (a , bs′ such-that m≽k) →
       proof  -- cool: a sub-proof
         relax m≽k (return a (≽-transitive n≽m m≽k) bs′)
       ≡⟨ reflexive ⟩ 
         relax m≽k (Maybe.return (a , bs′ such-that ≽-reflexive))
       ≡⟨ congruent (λ hole → Maybe.return (a , bs′ such-that hole)) (≽-right-identity m≽k) ⟩ 
         Maybe.return (a , bs′ such-that m≽k)
       ∎) ⟩
      (M) n≽m bs
    ∎

  --  associativity  : M >>= (λ a -> F a >>= G) ≡ (M >>= F) >>= G
  associativity : ∀ (n≽m : n ≽ m) (bs : Bits m) ->
      (M >>= (λ a -> F a >>= G)) n≽m bs ≡ ((M >>= F) >>= G) n≽m bs
  associativity {M = M} {F = F} {G = G} n≽m bs =
    proof
      (M >>= (λ a → F a >>= G)) n≽m bs
    ≡⟨ reflexive ⟩ 
      (M n≽m bs Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k ((F a >>= G) (≽-transitive n≽m m≽k) bs′))
    ≡⟨ reflexive ⟩
      (M n≽m bs Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (F a (≽-transitive n≽m m≽k) bs′ Maybe.>>= λ (b , bs″ such-that k≽j) -> relax k≽j (G b (≽-transitive (≽-transitive n≽m m≽k) k≽j) bs″)))
    ≡⟨ Maybe.>>=-congruent {m = M n≽m bs} (λ (a , bs′ such-that m≽k) →
       proof  -- cool: a sub-proof
         relax m≽k (F a (≽-transitive n≽m m≽k) bs′ Maybe.>>= λ (b , bs″ such-that k≽j) -> relax k≽j (G b (≽-transitive (≽-transitive n≽m m≽k) k≽j) bs″))
       ≡⟨ Maybe.>>=-map {m = F a (≽-transitive n≽m m≽k) bs′} ⟩ 
         (F a (≽-transitive n≽m m≽k) bs′ Maybe.>>= λ (b , bs″ such-that k≽j) -> relax m≽k (relax k≽j (G b (≽-transitive (≽-transitive n≽m m≽k) k≽j) bs″)))
       ≡⟨ Maybe.>>=-congruent {m = F a (≽-transitive n≽m m≽k) bs′} (λ (b , bs″ such-that k≽j) →
          proof  -- cool: a sub-sub-proof
            relax m≽k (relax k≽j (G b (≽-transitive (≽-transitive n≽m m≽k) k≽j) bs″))
          ≡⟨ symmetric (relax-preserves-composition m≽k k≽j _) ⟩ 
            relax (≽-transitive m≽k k≽j) (G b (≽-transitive (≽-transitive n≽m m≽k) k≽j) bs″)
          ≡⟨ congruent (λ hole → relax (≽-transitive m≽k k≽j) (G b hole bs″)) (≽-associative n≽m m≽k k≽j) ⟩ 
            relax (≽-transitive m≽k k≽j) (G b (≽-transitive n≽m (≽-transitive m≽k k≽j)) bs″)
          ∎) ⟩ 
         (F a (≽-transitive n≽m m≽k) bs′ Maybe.>>= λ (b , bs″ such-that k≽j) -> relax (≽-transitive m≽k k≽j) (G b (≽-transitive n≽m (≽-transitive m≽k k≽j)) bs″))
       ≡⟨ Maybe.map->>= {m = F a (≽-transitive n≽m m≽k) bs′} ⟩ 
         (relax m≽k (F a (≽-transitive n≽m m≽k) bs′) Maybe.>>= λ (b , bs′ such-that m≽j) -> relax m≽j (G b (≽-transitive n≽m m≽j) bs′))
       ∎) ⟩
      (M n≽m bs Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (F a (≽-transitive n≽m m≽k) bs′) Maybe.>>= λ (b , bs′ such-that m≽j) -> relax m≽j (G b (≽-transitive n≽m m≽j) bs′))
    ≡⟨ Maybe.associativity {m = M n≽m bs} ⟩ 
      ((M n≽m bs Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (F a (≽-transitive n≽m m≽k) bs′)) Maybe.>>= λ (b , bs′ such-that m≽j) -> relax m≽j (G b (≽-transitive n≽m m≽j) bs′))
    ≡⟨ reflexive ⟩ 
      ((M >>= F) n≽m bs Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (G a (≽-transitive n≽m m≽k) bs′))
    ≡⟨ reflexive ⟩ 
      ((M >>= F) >>= G) n≽m bs
    ∎
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Combinators specific to parsing. The case parser:

<<case>>
\begin{code}
O:_I:_ : (p q : IDecoder A i) -> IDecoder A i
(O: p I: q) (zero)      []      = undefined
(O: p I: q) (succ i≽j)  (𝟢 bs)  = relax succ-n≽n ((tighten succ-n≽n p) i≽j bs)
(O: p I: q) (succ i≽j)  (𝟣 bs)  = relax succ-n≽n ((tighten succ-n≽n q) i≽j bs)
\end{code}

The climax: the sole purpose of the set-up, indexed sets etc, is to be
able to define a version of case-analysis that allows for recursive
calls. O-rec:_I-rec:_ is a recursion scheme or a recursion
principle.

<<Case>>
\begin{code}
O-rec:_I-rec:_ :  (P Q : ∀ {i} -> IDecoder A i -> IDecoder A i)
                    -> (∀ {i} -> IDecoder A i)
O-rec: P I-rec: Q = recurse _ where
  recurse : ∀ i -> IDecoder _ i
  recurse i         (zero)      []      = undefined
  recurse (succ i)  (succ i≽j)  (𝟢 bs)  = relax succ-n≽n (P  (recurse i) i≽j bs)
  recurse (succ i)  (succ i≽j)  (𝟣 bs)  = relax succ-n≽n (Q  (recurse i) i≽j bs)
\end{code}

NB. The recursor is defined by induction on i.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Applicative interface (not used).

\begin{code}
pure : ∀ {i} -> A -> IDecoder A i
pure = return

infixl 4 _⍟_
_⍟_ : ∀ {i} -> IDecoder (A -> B) i -> (IDecoder A i -> IDecoder B i)
mf ⍟ ma = do f <- mf ; a <- ma ; return (f a)
\end{code}

-------------------------------------------------------------------------------

Combinators for encoding aka printing.

Encoders forms a monoid. Additionally, we provide encoders that emit a
single bit.  Encoders use an accumulating argument for efficiency, see
Hughes' efficient representation of lists.

<<Code>>
\begin{code}
Code : Set
Code = ∀ {i} -> Bits i -> Bits≽ i
\end{code}

<<empty>>
\begin{code}
ε : Code
ε = λ bs -> bs such-that ≼-reflexive
\end{code}

\begin{code}
infixr 10 _·_
\end{code}

<<concatenation>>
\begin{code}
_·_ : Code -> Code -> Code
f · g = λ bs ->  let bs′  such-that i≼j  = g bs   in
                 let bs″  such-that j≼k  = f bs′  in
                 bs″ such-that ≼-transitive i≼j j≼k
\end{code}

<<terminals>>
\begin{code}
O I : Code
O = λ bs -> 𝟢 bs such-that n≼succ-n
I = λ bs -> 𝟣 bs such-that n≼succ-n
\end{code}

The combinators form a monoid: · is associative with ε as its neutral element.

\begin{code}
·-left-identity : ∀ (s : Code) -> ∀ (bs : Bits n) ->
    (ε · s) bs ≡ (s) bs
·-left-identity s bs = congruent (λ hole -> bits (s bs) such-that hole) (≽-left-identity _)

·-right-identity : ∀ (s : Code) -> ∀ (bs : Bits n) ->
    (s · ε) bs ≡ (s) bs
·-right-identity s bs = congruent (bits (s bs) such-that_) (≽-right-identity _)

·-associative : ∀ (s t u : Code) -> ∀ (bs : Bits n) ->
    ((s · t) · u) bs ≡ (s · (t · u)) bs
·-associative s t u bs = congruent (bits (((s · t) · u) bs) such-that_) (≽-associative _ _ _)
\end{code}

-------------------------------------------------------------------------------

Proof combinators, a family of A-indexed relations or inference rules.

\begin{code}
infix 0 _⟨_⟩_
\end{code}

<<correct>>
\begin{code}
_⟨_⟩_ : Code -> A -> Decoder A -> Set
code ⟨ a ⟩ decode =
  ∀ k -> (bs : Bits k) -> ∀ i -> (i≽j : i ≽ size (code bs)) ->
    decode i≽j (bits (code bs)) ≡ Just (a , bs such-that above (code bs))
\end{code}

\begin{code}
private variable
  a : A
  s : Code
  p q : ∀ {i} -> IDecoder A i
\end{code}

<<empty-axiom>>
\begin{code}
ε-axiom : ε ⟨ a ⟩ return a
ε-axiom i bs j j≽k = ≡-reflexive
\end{code}

…, or, if you prefer equational proofs:

<<or-empty-axiom>>
\begin{code}
ε-axiom′ : ε ⟨ a ⟩ return a
ε-axiom′ {a = a} i bs j j≽k =
    proof
      return a j≽k (bits (ε bs))
    ≡⟨ ≡-reflexive ⟩
      Just (a , bs such-that above (ε bs))
    ∎
\end{code}

\begin{code}
·-rule : ∀ {s t : Code} {a : A} (p : ∀ {n} -> IDecoder A n) {b : B} (q : ∀ {n} -> A -> IDecoder B n) ->
      (s     ⟨ a ⟩ p      ) -> 
      (t     ⟨ b ⟩ q a    ) ->
      (s · t ⟨ b ⟩ p >>= q)
·-rule {s = s} {t = t} {a = a} p {b = b} q hyp-a hyp-b = λ k bs n n≽m →
-- NB. bits ((s · t) bs) ≡ bits (s (bits (t bs)))
    proof
      (p >>= q) n≽m (bits ((s · t) bs))
    ≡⟨ reflexive ⟩
      (p n≽m (bits ((s · t) bs)) Maybe.>>= λ (a , bs′ such-that m≽k) -> relax m≽k (q a (≽-transitive n≽m m≽k) bs′) )
    ≡⟨ Maybe.invoke-left-identity (hyp-a (size (t bs)) (bits (t bs)) n n≽m) ⟩
      relax (above (s (bits (t bs)))) (q a (≽-transitive n≽m (above (s (bits (t bs))))) (bits (t bs)))
    ≡⟨ congruent (relax (above (s (bits (t bs))))) (hyp-b k bs n (≽-transitive n≽m (above (s (bits (t bs)))))) ⟩ 
      relax (above (s (bits (t bs)))) (Just (b , bs such-that above (t bs)))
    ≡⟨ reflexive ⟩ 
      Just (b , bs such-that above ((s · t) bs))
    ∎
\end{code}

Rather annoyingly, we have to pass p and q as explicit arguments to
tell Agda how the monadic computation is split into two.

<<O-rule>>
\begin{code}
O-rule : (s ⟨ a ⟩ p) -> (O · s ⟨ a ⟩ (O: p I: q))
O-rule premise i bs (succ j) (succ j≽k) =
  ≡-congruent (relax succ-n≽n) (premise i bs (succ j) (≽-transitive succ-n≽n j≽k))
\end{code}

…, or, if you prefer equational proofs:

<<or-O-rule>>
\begin{code}
O-rule′ : (s ⟨ a ⟩ p) -> (O · s ⟨ a ⟩ (O: p I: q))
O-rule′ {s = s} {a = a} {p = p} {q = q} premise k bs (succ i) (succ i≽j) =
    proof
      relax succ-n≽n (p (≽-transitive succ-n≽n i≽j) (bits (s bs)))
    ≡⟨ ≡-congruent (relax succ-n≽n) (premise k bs (succ i) (≽-transitive succ-n≽n i≽j)) ⟩ 
      relax succ-n≽n (Just (a , bs such-that above (s bs)))
    ≡⟨ ≡-reflexive ⟩ 
      Just (a , bs such-that ≽-transitive succ-n≽n (above (s bs)))
    ≡⟨ ≡-reflexive ⟩ 
      Just (a , bs such-that above ((O · s) bs))
    ∎
\end{code}

\begin{code}
I-rule : ∀ {t : Code} {a : A} {f g : ∀ {n} -> IDecoder A n} ->
    (t ⟨ a ⟩ g) -> (I · t ⟨ a ⟩ (O: f I: g))
I-rule {t = t} {a = a} {f = f} {g = g} hyp k bs (succ n) (succ n≽m) =
    proof
      relax succ-n≽n (g (≽-transitive succ-n≽n n≽m) (bits (t bs)))
    ≡⟨ congruent (relax succ-n≽n) (hyp k bs (succ n) (≽-transitive succ-n≽n n≽m)) ⟩ 
      relax succ-n≽n (Just (a , bs such-that above (t bs)))
    ≡⟨ reflexive ⟩ 
      Just (a , bs such-that ≽-transitive succ-n≽n (above (t bs)))
    ≡⟨ reflexive ⟩ 
      Just (a , bs such-that above ((O · t) bs))
    ∎
\end{code}

\begin{code}
O-Rule : ∀ {s : Code} {a : A} {F G : ∀ {n} -> IDecoder A n -> IDecoder A n} ->
    (s ⟨ a ⟩ F (O-rec: F I-rec: G)) -> 
    (O · s ⟨ a ⟩ (O-rec: F I-rec: G))
O-Rule {s = s} {a = a} {F = F} {G = G} hyp k bs (succ n) (succ n≽m) =
    proof
      relax succ-n≽n (F (O-rec: F I-rec: G) n≽m (bits (s bs)))
    ≡⟨ congruent (relax succ-n≽n) (hyp k bs n n≽m) ⟩ 
      relax succ-n≽n (Just (a , bs such-that above (s bs)))
    ≡⟨ reflexive ⟩ 
      Just (a , bs such-that ≽-transitive succ-n≽n (above (s bs)))
    ≡⟨ reflexive ⟩ 
    Just (a , bs such-that above ((O · s) bs))
  ∎
\end{code}

\begin{code}
I-Rule : ∀ {s : Code} {a : A} {F G : ∀ {n} -> IDecoder A n -> IDecoder A n} ->
    (s ⟨ a ⟩ G (O-rec: F I-rec: G)) -> 
    (I · s ⟨ a ⟩ (O-rec: F I-rec: G))
I-Rule {s = s} {a = a} {F = F} {G = G} hyp k bs (succ n) (succ n≽m) =
    proof
      relax succ-n≽n (G (O-rec: F I-rec: G) n≽m (bits (s bs)))
    ≡⟨ congruent (relax succ-n≽n) (hyp k bs n n≽m) ⟩ 
      relax succ-n≽n (Just (a , bs such-that above (s bs)))
    ≡⟨ reflexive ⟩ 
      Just (a , bs such-that ≽-transitive succ-n≽n (above (s bs)))
    ≡⟨ reflexive ⟩ 
    Just (a , bs such-that above ((O · s) bs))
  ∎
\end{code}

-------------------------------------------------------------------------------
