A „type class“ bundling parsers, printers, and proofs.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Serialise.Indexed.Codec where

import VFP.Maybe as Maybe
open import VFP.Serialise.Indexed public
open import VFP.Natural.Non-Strict-Order

private variable
  A B : Set
\end{code}

-------------------------------------------------------------------------------

In some sense, a simplified version of Haskell's Show and Read
classes, except for the correctness proof.

<<Codec>>
\begin{code}
record Codec (A : Set) : Set where
  field  encode   : A -> Code
         decode   : Decoder A
         correct  : ∀ (a : A) -> encode a ⟨ a ⟩ decode
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

A mappping function …

\begin{code}
open import VFP.Bijection.Properties
\end{code}

<<Map-encode-decode>>
\begin{code}
module Codec-Iso (iso : A ≅ B) (𝔸 : Codec A) where
  encode : B -> Code
  encode b = Codec.encode 𝔸 (fro iso b)

  decode : Decoder B
  decode =  do a <- Codec.decode 𝔸 ; return (to iso a)
\end{code}

\begin{code}
  open import VFP.Equality.Reasoning
  correct : ∀ (b : B) -> encode b ⟨ b ⟩ decode
  correct b k bs n n≽m =
    proof
      decode n≽m (bits (encode b bs))
    ≡⟨ reflexive ⟩
      (Codec.decode 𝔸 >>= λ a -> return (to iso a)) n≽m (bits (encode b bs))
    ≡⟨ Maybe.invoke (Codec.correct 𝔸 (fro iso b) k bs n n≽m) ⟩
      Maybe.Just (to iso (fro iso b) , raise (above (encode b bs)) (bs such-that ≽-reflexive))
    ≡⟨ congruent (λ hole -> Maybe.Just (hole , raise (above (encode b bs)) (bs such-that ≽-reflexive))) (to∘fro≡id iso b) ⟩ 
      Maybe.Just (b , raise (above (encode b bs)) (bs such-that ≽-reflexive))
    ≡⟨ reflexive ⟩
      Maybe.Just (b , bs such-that ≽-transitive (above (encode b bs)) ≽-reflexive)
    ≡⟨ congruent (λ hole → Maybe.Just (b , bs such-that hole)) (≽-right-identity _) ⟩ 
      Maybe.Just (b , bs such-that above (encode b bs))
    ∎
\end{code}

<<Map-record>>
\begin{code}
𝕄𝕒𝕡 : (A ≅ B) -> (Codec A -> Codec B)
\end{code}

\begin{code}
𝕄𝕒𝕡 iso 𝔸 = record { Codec-Iso iso 𝔸 }
\end{code}

-------------------------------------------------------------------------------

Encoders and decoders for some (standard) types.

\begin{code}
open import VFP.Boolean

module Codec-Boolean where
  encode : Boolean -> Code
  encode false = O · ε
  encode true  = I · ε

  decode : Decoder Boolean
  decode = O: return false I: return true

  correct : ∀ a -> encode a ⟨ a ⟩ decode
  correct false = O-rule ε-axiom
  correct true  = I-rule ε-axiom

𝔹𝕠𝕠𝕝𝕖𝕒𝕟 : Codec Boolean
𝔹𝕠𝕠𝕝𝕖𝕒𝕟 = record { Codec-Boolean }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
open import VFP.Natural
\end{code}

<<Natural-Codec>>
\begin{code}
module Codec-Natural where
  encode : Natural -> Code
  encode (zero)    = O · ε
  encode (succ n)  = I · encode n · ε

  decode : Decoder Natural
  decode =  O-rec:  (λ natural → return zero)
            I-rec:  (λ natural → do n <- natural ; return (succ n))

  correct : ∀ (n : Natural)-> encode n ⟨ n ⟩ decode
  correct (zero)    = O-Rule ε-axiom
  correct (succ n)  = I-Rule  (·-rule decode (λ n′ → return (succ n′)) (correct n)
                              (ε-axiom))

ℕ𝕒𝕥𝕦𝕣𝕒𝕝 : Codec Natural
ℕ𝕒𝕥𝕦𝕣𝕒𝕝 = record { Codec-Natural }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
open import VFP.Ternary hiding (_𝟘)

module Codec-Ternary where
  encode : Ternary -> Code
  encode (𝟘t)  = O · O · ε             -- NB. the ε is necessary to ensure that the proof combinators are applicable
  encode (n 𝟙) = O · I · encode n · ε  -- General rule: if the decoder ends with return, then the encoder has to end with ε.
  encode (n 𝟚) = I · O · encode n · ε  -- This is because · is no longer associative on the nose (i.e. definitionally)!
  encode (n 𝟛) = I · I · encode n · ε

  decode : Decoder Ternary
  decode = O-rec: (λ recurse → O: (return 𝟘t)
                                I: (do n <- recurse ; return (n 𝟙)))
           I-rec: (λ recurse → O: (do n <- recurse ; return (n 𝟚))
                                I: (do n <- recurse ; return (n 𝟛)))
                                
  correct : ∀ (n : Ternary) -> encode n ⟨ n ⟩ decode
  correct (𝟘t)  = O-Rule (O-rule ε-axiom)
  correct (n 𝟙) = O-Rule (I-rule (·-rule decode (λ n -> return (n 𝟙)) (correct n) ε-axiom))
  correct (n 𝟚) = I-Rule (O-rule (·-rule decode (λ n -> return (n 𝟚)) (correct n) ε-axiom))
  correct (n 𝟛) = I-Rule (I-rule (·-rule decode (λ n -> return (n 𝟛)) (correct n) ε-axiom))

𝕋𝕖𝕟𝕒𝕣𝕪 : Codec Ternary
𝕋𝕖𝕟𝕒𝕣𝕪 = record { Codec-Ternary }
\end{code}

In general, n constructors can be encoded with ⌈ lg n ⌉ bits.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<Bush>>
\begin{code}
data Bush : Set where
  Leaf  : Bush
  Fork  : Bush -> Bush -> Bush
\end{code}
  
<<Codec-Bush>>
\begin{code}
module Codec-Bush where
\end{code}

<<Bush-encode>>
\begin{code}
  encode : Bush -> Code
  encode (Leaf)      = O  · ε
  encode (Fork t u)  = I  · encode t · encode u · ε
\end{code}

\begin{spec}
  decode′ : Decoder Bush
\end{spec}

<<Bush-decode′>>
\begin{spec}
  {-# TERMINATING #-}
  decode′ =  O:  return Leaf
             I:  do t <- decode′ ; u <- decode′ ; return (Fork t u)
\end{spec}

<<Bush-decode>>
\begin{code}
  decode : Decoder Bush
  decode =  O-rec:  (λ bush → return Leaf)
            I-rec:  (λ bush → do t <- bush ; u <- bush ; return (Fork t u))
\end{code}

<<Bush-correct>>
\begin{code}
  correct : ∀ (t : Bush) -> encode t ⟨ t ⟩ decode
  correct (Leaf)      = O-Rule  ε-axiom
  correct (Fork t u)  =
    I-Rule  (·-rule decode (λ t′ -> do u′ <- decode ;  return (Fork t′  u′)) (correct t)
            (·-rule decode (λ u′ ->                    return (Fork t   u′)) (correct u) 
            (ε-axiom)))
\end{code}

<<Bush-Codec>>
\begin{code}
𝔹𝕦𝕤𝕙 : Codec Bush
𝔹𝕦𝕤𝕙 = record { Codec-Bush }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Serializing elements of dependent types. Vectors: no bits required!

\begin{code}
open import VFP.Vector
\end{code}

<<Vector-encode-decode>>
\begin{code}
module Codec-Vector {A : Set} (𝔸 : Codec A) where
  encode : {i : Natural} -> Vector A i -> Code
  encode ⟨⟩        = ε
  encode (a , as)  = Codec.encode 𝔸 a · encode as · ε

  decode : (i : Natural) -> Decoder (Vector A i)
  decode (zero)    = return ⟨⟩
  decode (succ i)  = do a ← Codec.decode 𝔸 ; as ← decode i ; return (a , as)
\end{code}

\begin{code}
  correct : (i : Natural) -> ∀ (v : Vector A i) -> encode v ⟨ v ⟩ decode i
  correct (zero)   ⟨⟩        = ε-axiom
  correct (succ i) (a , as)  = ·-rule (Codec.decode 𝔸) (λ a -> do as ← decode i ; return (a , as)) (Codec.correct 𝔸 a)
                               (·-rule (decode i)                        (λ as -> return (a , as)) (correct i as)
                               ε-axiom)
\end{code}

<<Vector-record>>
\begin{code}
𝕍𝕖𝕔𝕥𝕠𝕣 : {A : Set} -> Codec A -> (i : Natural) -> Codec (Vector A i)
\end{code}

\begin{code}
𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 i = record
  { encode  = Codec-Vector.encode  𝔸
  ; decode  = Codec-Vector.decode  𝔸 i
  ; correct = Codec-Vector.correct 𝔸 i}
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<List>>
\begin{code}
open import VFP.List using (List ; [] ; _∷_)
\end{code}

<<List-encode-decode>>
\begin{code}
module Codec-List {A : Set} (𝔸 : Codec A) where
  encode : List A -> Code
  encode []        = O · ε
  encode (a ∷ as)  = I · Codec.encode 𝔸 a · encode as · ε

  decode : Decoder (List A)
  decode =  O-rec:  (λ list → return [])
            I-rec:  (λ list → do a ← Codec.decode 𝔸 ; as ← list ; return (a ∷ as))
\end{code}

\begin{code}
  correct : ∀ (as : List A) -> encode as ⟨ as ⟩ decode
  correct []       = O-Rule ε-axiom
  correct (a ∷ as) = I-Rule (·-rule (Codec.decode 𝔸) (λ a′ -> do as′ ← decode ; return (a′ ∷ as′)) (Codec.correct 𝔸 a)
                              (·-rule decode                          (λ as′ -> return (a ∷ as′)) (correct as)
                              ε-axiom))
\end{code}

<<List-record>>
\begin{code}
𝕃𝕚𝕤𝕥 : {A : Set} -> Codec A -> Codec (List A)
\end{code}

\begin{code}
𝕃𝕚𝕤𝕥 𝔸 = record { Codec-List 𝔸 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Serialiser for dependent sums/products aka Sigma types.

<<Sigma-encode-decode>>
\begin{code}
module Codec-Σ  {A : Set} (𝔸 : Codec A) 
                  {B : A → Set} (𝔹 : (a : A) -> Codec (B a)) where
  encode : Σ A B -> Code
  encode (a , b) = Codec.encode 𝔸 a · Codec.encode (𝔹 a) b · ε

  decode : Decoder (Σ A B)
  decode = do a <- Codec.decode 𝔸 ; b <- Codec.decode (𝔹 a) ; return (a , b)
\end{code}

\begin{code}
  correct : ∀ (p : Σ A B) -> encode p ⟨ p ⟩ decode
  correct (a , b) = ·-rule (Codec.decode 𝔸)     (λ a -> do b <- Codec.decode (𝔹 a) ; return (a , b)) (Codec.correct 𝔸 a) (
                    ·-rule (Codec.decode (𝔹 a)) (                      λ b -> return (a , b)) (Codec.correct (𝔹 a) b) (
                    ε-axiom))
\end{code}

<<Sigma-record>>
\begin{code}
𝕊𝕚𝕘𝕞𝕒 :  {A : Set} -> Codec A -> 
            {B : A → Set} -> ((a : A) -> Codec (B a)) -> 
              Codec (Σ A B)
\end{code}

\begin{code}
𝕊𝕚𝕘𝕞𝕒 𝔸 𝔹 = record { Codec-Σ 𝔸 𝔹 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<List-prime>>
\begin{code}
List′ : Set -> Set
List′ A = Σ Natural (Vector A)
\end{code}

Ad-hoc encoder and decoder (more efficient version further below).

\begin{code}
module Codec-List′ {A : Set} (𝔸 : Codec A) where
  encode : List′ A -> Code
  encode (n , as) = Codec.encode ℕ𝕒𝕥𝕦𝕣𝕒𝕝 n · Codec.encode (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) as · ε

  decode : Decoder (List′ A)
  decode = do n <- Codec.decode ℕ𝕒𝕥𝕦𝕣𝕒𝕝 ; as <- Codec.decode (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) ; return (n , as)

  correct : ∀ (x : List′ A) -> encode x ⟨ x ⟩ decode
  correct (n , as) = ·-rule (Codec.decode ℕ𝕒𝕥𝕦𝕣𝕒𝕝)     (λ n -> do as <- Codec.decode (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) ; return (n , as)) (Codec.correct ℕ𝕒𝕥𝕦𝕣𝕒𝕝 n) (
                    ·-rule (Codec.decode (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n)) (                      λ as -> return (n , as)) (Codec.correct (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸 n) as) (
                    ε-axiom))

ad-hoc-𝕃𝕚𝕤𝕥′ : {A : Set} -> Codec A -> Codec (List′ A)
ad-hoc-𝕃𝕚𝕤𝕥′ 𝔸 = record { Codec-List′ 𝔸 }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
Ternary≅Natural = ≅-symmetric Natural≅Ternary
\end{code}

<<Ternary>>
\begin{code}
ℕ𝕒𝕥𝕦𝕣𝕒𝕝′ : Codec Natural
ℕ𝕒𝕥𝕦𝕣𝕒𝕝′ = 𝕄𝕒𝕡 Ternary≅Natural 𝕋𝕖𝕟𝕒𝕣𝕪
\end{code}

<<List-prime-record>>
\begin{code}
𝕃𝕚𝕤𝕥′ : {A : Set} -> Codec A -> Codec (List′ A)
𝕃𝕚𝕤𝕥′ 𝔸 = 𝕊𝕚𝕘𝕞𝕒 (𝕄𝕒𝕡 Ternary≅Natural 𝕋𝕖𝕟𝕒𝕣𝕪) (𝕍𝕖𝕔𝕥𝕠𝕣 𝔸)
\end{code}

Ctrl-C Ctrl-N: Codec.encode ℕ𝕒𝕥𝕦𝕣𝕒𝕝′ 4711

-------------------------------------------------------------------------------
