Huffman encoding (we assume that the code tree is given to us).

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --safe #-}

module VFP.Serialise.Indexed.Huffman where

open import VFP.Equality.Properties
open import VFP.Logic
open import VFP.Serialise.Indexed.Codec public

private variable
  A B Alphabet : Set
\end{code}

-------------------------------------------------------------------------------

Serialiser for Huffman code trees.

<<Code-Tree>>
\begin{code}
data Code-Tree (Alphabet : Set) : Set where
  ⌊_⌋ : Alphabet -> Code-Tree Alphabet
  _⋏_ : Code-Tree Alphabet -> Code-Tree Alphabet -> Code-Tree Alphabet
\end{code}

<<example>>
\begin{spec}
code =  ((⌊ 'h' ⌋ ⋏ ⌊ 't' ⌋) ⋏ ⌊ 'e' ⌋)
          ⋏ ((⌊ 'i' ⌋ ⋏ ⌊ 'n' ⌋) ⋏ ((⌊ 'v' ⌋ ⋏ ⌊ 'g' ⌋) ⋏ ⌊ blank ⌋))
\end{spec}

<<Codec-Code-Tree-encode-decode>>
\begin{code}
module Codec-Code-Tree {A : Set} (𝔸 : Codec A) where
  encode : Code-Tree A -> Code
  encode ⌊ a ⌋    = O · Codec.encode 𝔸 a · ε
  encode (t ⋏ u)  = I · encode t · encode u · ε

  decode : Decoder (Code-Tree A)
  decode =  O-rec:  (λ tree -> do a ← Codec.decode 𝔸 ; return ⌊ a ⌋)
            I-rec:  (λ tree -> do t ← tree ; u ← tree ; return (t ⋏ u))
\end{code}

\begin{code}
  correct : ∀ (t : Code-Tree A) -> encode t ⟨ t ⟩ decode
  correct ⌊ a ⌋    = O-Rule (·-rule (Codec.decode 𝔸) ((λ a′ -> return ⌊ a′ ⌋)) (Codec.correct 𝔸 a)
                            ε-axiom)
  correct (t ⋏ u)  = I-Rule (·-rule decode (λ t′ -> do u′ ← decode ;  return (t′  ⋏ u′)) (correct t)
                            (·-rule decode (λ u′ ->                   return (t   ⋏ u′)) (correct u) 
                            (ε-axiom)))
\end{code}

<<Codec-Code-Tree-record>>
\begin{code}
ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 : {A : Set} -> Codec A -> Codec (Code-Tree A)
\end{code}

\begin{code}
ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝔸 = record { Codec-Code-Tree 𝔸 } 
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Classical Huffman encoding.

Adapted code from the original submission.

Existence: evidence/witness is a path into the tree.

<<exists>>
\begin{code}
data exists (P : Alphabet -> Set) : (Code-Tree Alphabet -> Set) where
  here   : ∀ {c}    -> P c         -> exists P ⌊ c ⌋
  left   : ∀ {t u}  -> exists P t  -> exists P (t ⋏ u)
  right  : ∀ {t u}  -> exists P u  -> exists P (t ⋏ u)
\end{code}

<<i-in-code>>
\begin{spec}
i-am-in  :  exists (λ c′ -> 'i' ≡ c′) code
i-am-in  =  right (left (left (here reflexive)))
\end{spec}

<<epsilon>>
\begin{code}
_⊆_ : (Alphabet : Set) -> Code-Tree Alphabet -> Set
Alphabet ⊆ t = exists (λ c -> c ≡ c) t
\end{code}

huffman-encode: driven by the path; huffman-decode: driven by bit
string and the tree.

<<Compress-encode>>
\begin{code}
module Compress where
  encode : ∀ {t : Code-Tree Alphabet} -> Alphabet ⊆ t -> Code
  encode (here   e)  = ε
  encode (left   p)  = O · encode p · ε
  encode (right  p)  = I · encode p · ε
\end{code}

<<Compress-decode>>
\begin{code}
  decode : ∀ (t : Code-Tree Alphabet) -> Decoder (Alphabet ⊆ t)
  decode ⌊ c ⌋   =  return (here reflexive)
  decode (t ⋏ u) =  O:  (do p <- decode t  ; return (left   p))
                    I:  (do p <- decode u  ; return (right  p))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

New code.

Using _∈_ instead of exists (above the character lives in the here
constructor i.e. at the rear of the path; below the character is the
first component of a dependent pair i.e. at the front of the path).

\begin{code}
infix 6 _∈_
private variable
  t u : Code-Tree Alphabet
\end{code}

<<exists>>
\begin{code}
data _∈_ (c : Alphabet) : Code-Tree Alphabet -> Set where
  here   : c ∈ ⌊ c ⌋
  left   : c ∈ t  -> c ∈ (t ⋏ u)
  right  : c ∈ u  -> c ∈ (t ⋏ u)
\end{code}

  left   : ∀ {t u}  -> c ∈ t  -> c ∈ (t ⋏ u)
  right  : ∀ {t u}  -> c ∈ u  -> c ∈ (t ⋏ u)

<<i-in-code>>
\begin{spec}
i-am-in  :  'i' ∈ code
i-am-in  =  right (left (left here))
\end{spec}

<<epsilon>>
\begin{code}
In : {Alphabet : Set} -> Code-Tree Alphabet -> Set
In {Alphabet} t = Σ Alphabet (λ c -> c ∈ t)
\end{code}

huffman-encode: driven by the path; huffman-decode: driven by bit
string and the tree.

<<Compress-encode>>
\begin{code}
module Codec-In where
  encode : ∀ {t : Code-Tree Alphabet} -> In t -> Code
  encode (c , here    )  = ε
  encode (c , left   p)  = O · encode (c , p) · ε
  encode (c , right  p)  = I · encode (c , p) · ε
\end{code}

<<Compress-decode>>
\begin{code}
  decode : ∀ (t : Code-Tree Alphabet) -> Decoder (In t)
  decode ⌊ c ⌋   =  return (c , here)
  decode (t ⋏ u) =  O:  (do (c , p) <- decode t  ; return (c , left   p))
                    I:  (do (c , p) <- decode u  ; return (c , right  p))
\end{code}

<<Compress-correct>>
\begin{code}
  correct : ∀ {t : Code-Tree Alphabet} -> ∀ (c : In t) -> encode c ⟨ c ⟩ decode t
  correct (c , here)      =  ε-axiom
  correct (c , left   p)  =  O-rule (·-rule (decode _) (λ (c′ , p′) ->
                             return (c′ , left p′)) (correct (c , p)) ε-axiom)
  correct (c , right  p)  =  I-rule (·-rule (decode _) (λ (c′ , p′) ->
                             return (c′ , right p′)) (correct (c , p)) ε-axiom)
\end{code}

<<Compress-record>>
\begin{code}
𝕀𝕟 : ∀ (t : Code-Tree Alphabet) -> Codec (In t)
\end{code}

\begin{code}
𝕀𝕟 t = record
  { encode  = Codec-In.encode
  ; decode  = Codec-In.decode t
  ; correct = Codec-In.correct }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Ye another alternative design.

\begin{code}
data Path {Alphabet : Set} : Code-Tree Alphabet -> Set where
  here   : ∀ {c }               -> Path ⌊ c ⌋
  left   : ∀ {t u }  -> Path t  -> Path (t ⋏ u)
  right  : ∀ {t u }  -> Path u  -> Path (t ⋏ u)
\end{code}

\begin{code}
module Codec-Path where
  encode : ∀ {t : Code-Tree Alphabet} -> Path t -> Code
  encode (here    )  = ε
  encode (left   p)  = O · encode p · ε
  encode (right  p)  = I · encode p · ε

  decode : ∀ (t : Code-Tree Alphabet) -> Decoder (Path t)
  decode ⌊ c ⌋   =  return here
  decode (t ⋏ u) =  O:  (do p <- decode t  ; return (left   p))
                    I:  (do p <- decode u  ; return (right  p))

  correct : ∀ {t : Code-Tree Alphabet} -> ∀ (c : Path t) -> encode c ⟨ c ⟩ decode t
  correct (here)            = ε-axiom  -- needs Axiom K???
  correct (left   p)        = O-rule (·-rule (decode _) (λ p -> return (left  p)) (correct p) ε-axiom)
  correct (right  p)        = I-rule (·-rule (decode _) (λ p -> return (right p)) (correct p) ε-axiom)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Encoding a list of characters paired with its code tree.

<<Huffman>>
\begin{code}
Huffman : Set -> Set
Huffman Alphabet = Σ (Code-Tree Alphabet) (λ t -> List′ (In t))
\end{code}

Ad-hoc encoder and decoder.

\begin{code}
module Codec-Huffman {Alphabet : Set} (𝕔𝕙𝕒𝕣 : Codec Alphabet) where
  encode : Huffman Alphabet -> Code
  encode (t , as) = Codec.encode (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) t · Codec.encode (𝕃𝕚𝕤𝕥′ (𝕀𝕟 t)) as · ε

  decode : Decoder (Huffman Alphabet)
  decode = do t <- Codec.decode (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) ; as <- Codec.decode (𝕃𝕚𝕤𝕥′ (𝕀𝕟 t)) ; return (t , as)

  correct : ∀ (x : Huffman Alphabet) -> encode x ⟨ x ⟩ decode
  correct (t , as) = ·-rule (Codec.decode (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣)) (λ t -> do as <- Codec.decode (𝕃𝕚𝕤𝕥′ (𝕀𝕟 t)) ; return (t , as) ) (Codec.correct (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) t)
                     (·-rule (Codec.decode (𝕃𝕚𝕤𝕥′ (𝕀𝕟 t))) (λ as -> return (t , as)) (Codec.correct (𝕃𝕚𝕤𝕥′ (𝕀𝕟 t)) as) ε-axiom)

ℍ𝕦𝕗𝕗𝕞𝕒𝕟′ : {Alphabet : Set} -> Codec Alphabet -> Codec (Huffman Alphabet)
ℍ𝕦𝕗𝕗𝕞𝕒𝕟′ 𝕔𝕙𝕒𝕣 = record { Codec-Huffman 𝕔𝕙𝕒𝕣 }
\end{code}

Or a tad shorter using 𝕊𝕚𝕘𝕞𝕒.

<<Huffman-record>>
\begin{code}
ℍ𝕦𝕗𝕗𝕞𝕒𝕟 : {Alphabet : Set} -> Codec Alphabet -> Codec (Huffman Alphabet)
ℍ𝕦𝕗𝕗𝕞𝕒𝕟 𝕔𝕙𝕒𝕣 = 𝕊𝕚𝕘𝕞𝕒 (ℂ𝕠𝕕𝕖-𝕋𝕣𝕖𝕖 𝕔𝕙𝕒𝕣) (λ t -> 𝕃𝕚𝕤𝕥′ (𝕀𝕟 t))
\end{code}

Modularity: lists can be easily replaced by other collection types.

-------------------------------------------------------------------------------
